#!/usr/bin/env python
# -*- coding: utf-8 -*-

import bisect
import json
import logging

# Built-in imports
import os

# 3rd party imports
import pandas as pd

# Local imports
from pyrfu.mms.list_files_ancillary import list_files_ancillary
from pyrfu.pyrf.extend_tint import extend_tint
from pyrfu.pyrf.iso86012datetime import iso86012datetime

__author__ = "Louis Richard"
__email__ = "louisr@irfu.se"
__copyright__ = "Copyright 2020-2023"
__license__ = "MIT"
__version__ = "2.4.2"
__status__ = "Prototype"

logging.captureWarnings(True)
logging.basicConfig(
    format="[%(asctime)s] %(levelname)s: %(message)s",
    datefmt="%d-%b-%y %H:%M:%S",
    level=logging.INFO,
)


def load_ancillary(
    product,
    tint,
    mms_id,
    verbose: bool = True,
    data_path: str = "",
):
    r"""Loads ancillary data.

    Parameters
    ----------
    product : {"predatt", "predeph", "defatt", "defeph"}
        Ancillary type.
    tint : list of str
        Time interval
    mms_id : str or int
        Spacecraft index
    verbose : bool, Optional
        Set to True to follow the loading. Default is True
    data_path : str, Optional
        Path of MMS data. If None use `pyrfu.mms.mms_config.py`

    Returns
    -------
    out : xarray.Dataset
        Time series of the ancillary data

    """

    # Get path of files in interval
    tint_long = extend_tint(tint, [-86400, 86400])
    files_names = list_files_ancillary(tint_long, mms_id, product, data_path)

    # Convert time interval to datetime
    tint = iso86012datetime(tint)

    # Read length of header and columns names from .json file
    # Root path
    pkg_path = os.path.dirname(os.path.abspath(__file__))

    with open(os.sep.join([pkg_path, "ancillary.json"]), "r", encoding="utf-8") as file:
        anc_dict = json.load(file)

    if verbose:
        logging.info("Loading ancillary %s files...", product)

    data_frame_dict = {}

    for i, file in enumerate(files_names):
        rows = pd.read_csv(
            file,
            # delim_whitespace=True,
            sep=r"\s+",
            header=None,
            skiprows=anc_dict[product]["header"],
        )

        # Remove footer
        rows = rows[:][:-1]

        # Convert time
        fmt = anc_dict[product]["time_format"]
        rows[0] = pd.to_datetime(rows[0], format=fmt)

        start_idx = bisect.bisect_left(rows[0][:], tint[0])
        end_idx = bisect.bisect_left(rows[0][:], tint[1])
        rows.columns = anc_dict[product]["columns_names"]

        data_frame_dict[i] = rows[:][start_idx:end_idx]

    data_frame = data_frame_dict[0]

    for k in list(data_frame_dict.keys())[1:]:
        data_frame = pd.concat([data_frame, data_frame_dict[k]], ignore_index=True)

    data_frame = data_frame.sort_values(by="time").set_index(["time"])
    dataset_out = data_frame.to_xarray()

    dataset_out = dataset_out.assign_coords(time=dataset_out.time.astype("<M8[ns]"))

    return dataset_out



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# Built-in imports
from typing import Mapping, Optional, Union

# 3rd party imports
import numpy as np
import xarray as xr
from numpy.typing import NDArray
from xarray.core.dataarray import DataArray

__author__ = "Louis Richard"
__email__ = "louisr@irfu.se"
__copyright__ = "Copyright 2020-2023"
__license__ = "MIT"
__version__ = "2.4.2"
__status__ = "Prototype"

NDArrayFloats = NDArray[Union[np.float32, np.float64]]


def ts_spectr(
    time: NDArray[np.datetime64],
    energy: NDArrayFloats,
    data: NDArrayFloats,
    comp_name: Optional[str] = None,
    attrs: Optional[Mapping[str, object]] = None,
) -> DataArray:
    r"""Create a time series containing a spectrum

    Parameters
    ----------
    time : numpy.ndarray
        Array of times.
    energy : numpy.ndarray
        Y value of the spectrum (energies, frequencies, etc.)
    data : numpy.ndarray
        Data of the spectrum.
    attrs : dict, Optional
        Attributes of the data list.

    Returns
    -------
    out : xarray.DataArray
        Time series of a spectrum

    """

    # Check input type
    if not isinstance(time, np.ndarray):
        raise TypeError("time must be a numpy.ndarray")

    if not isinstance(energy, np.ndarray):
        raise TypeError("time must be a numpy.ndarray")

    if not isinstance(data, np.ndarray):
        raise TypeError("data must be a numpy.ndarray")

    # Check input shape must be (n, m, )
    if data.ndim != 2:
        raise ValueError("Input must be a spectrum")

    if len(time) != data.shape[0]:
        raise ValueError("Shape mismatch. Time and data must have the same length")

    if len(energy) != data.shape[1]:
        raise ValueError("Shape mismatch. Energy and data must have the same length")

    if comp_name is None:
        comp_name = "energy"

    if attrs is None or not isinstance(attrs, dict):
        attrs = {}

    out: DataArray = xr.DataArray(
        data, coords=[time, energy], dims=["time", comp_name], attrs=attrs
    )
    out.attrs["TENSOR_ORDER"] = 0

    return out



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# 3rd party imports
import xarray as xr

from .db_get_ts import db_get_ts
from .db_get_variable import db_get_variable

# Local imports
from .list_files import list_files

__author__ = "Louis Richard"
__email__ = "louisr@irfu.se"
__copyright__ = "Copyright 2020-2023"
__license__ = "MIT"
__version__ = "2.4.2"
__status__ = "Prototype"


def get_eis_allt(
    tar_var,
    tint,
    mms_id,
    verbose: bool = True,
    data_path: str = "",
):
    r"""Read energy spectrum of the selected specie in the selected energy
    range for all telescopes.

    Parameters
    ----------
    tar_var : str
        Key of the target variable like
        {data_unit}_{dtype}_{specie}_{data_rate}_{data_lvl}.
    tint : list of str
        Time interval.
    mms_id : int or float or str
        Index of the spacecraft.
    verbose : bool, Optional
        Set to True to follow the loading. Default is True.
    data_path : str, Optional
        Path of MMS data.

    Returns
    -------
    out : xarray.Dataset
        Dataset containing the energy spectrum of the 6 telescopes of the
        Energy Ion Spectrometer.

    Examples
    --------
    >>> from pyrfu import mms

    Define time interval

    >>> tint_brst = ["2017-07-23T16:54:24.000", "2017-07-23T17:00:00.000"]

    Read proton energy spectrum for all EIS telescopes

    >>> eis_allt = mms.get_eis_allt("Flux_extof_proton_srvy_l2", tint_brst, 2)

    """

    # Convert mms_id to integer
    mms_id = int(mms_id)

    data_unit, data_type, specie, data_rate, data_lvl = tar_var.split("_")

    if tint[0] < "2020-01-01":
        pref = f"mms{mms_id:d}_epd_eis"
    else:
        pref = f"mms{mms_id:d}_epd_eis_{data_rate}_{data_lvl}"

    var = {
        "mms_id": mms_id,
        "inst": "epd-eis",
        "dtype": data_type,
        "tmmode": data_rate,
        "lev": data_lvl,
        "specie": specie,
        "data_path": data_path,
    }

    # EIS includes the version of the files in the cdfname need to read it
    # before.
    files = list_files(tint, mms_id, var, data_path=data_path)

    if not files:
        raise FileNotFoundError("no files for these inputs!!")

    file_version = int(files[0].split("_")[-1][1])
    var["version"] = file_version

    pref = f"{pref}_{data_rate}_{data_lvl}_{data_type}"
    if int(files[0].split("_")[-1][3]) >= 1:
        if specie == "alpha":
            specie = "helium"
    else:
        if data_rate == "brst":
            pref = f"{pref}_{data_rate}_{data_type}"
        else:
            pref = f"{pref}_{data_type}"

    if data_unit.lower() in ["flux", "counts", "cps"]:
        suf = f"{specie}_P{file_version:d}_{data_unit.lower()}_t"
    else:
        raise ValueError("Invalid data unit")

    # Name of the data containing index of the probe, instrument, data rate,
    # data level and data type if needed
    dset_name = (
        f"mms{var['mms_id']:d}_{var['inst']}_{var['tmmode']}"
        f"_{var['lev']}_{var['dtype']}"
    )

    # Names of the energy spectra in the CDF (one for each telescope)
    cdfnames = [f"{pref}_{suf}{t:d}" for t in range(6)]

    spin_nums = db_get_ts(dset_name, f"{pref}_spin", tint, data_path=data_path)
    sectors = db_get_ts(dset_name, f"{pref}_sector", tint, data_path=data_path)

    e_minus = db_get_variable(
        dset_name,
        f"{pref}_{specie}_t0_energy_dminus",
        tint,
        verbose=verbose,
        data_path=data_path,
    )

    e_plus = db_get_variable(
        dset_name,
        f"{pref}_{specie}_t0_energy_dplus",
        tint,
        verbose=verbose,
        data_path=data_path,
    )

    outdict = {"spin": spin_nums, "sector": sectors}

    for i, cdfname in enumerate(cdfnames):
        scope_key = f"t{i:d}"

        outdict[scope_key] = db_get_ts(
            dset_name,
            cdfname,
            tint,
            verbose=verbose,
            data_path=data_path,
        )
        outdict[scope_key] = outdict[scope_key].rename(
            {"time": "time", "Energy": "energy"},
        )

        outdict[f"look_{scope_key}"] = db_get_ts(
            dset_name,
            f"{pref}_look_{scope_key}",
            tint,
            verbose=verbose,
            data_path=data_path,
        )

    e_plus = e_plus.assign_coords(x=outdict["t0"].energy.data)
    e_minus = e_minus.assign_coords(x=outdict["t0"].energy.data)
    e_plus = e_plus.rename({"x": "energy"})
    e_minus = e_minus.rename({"x": "energy"})

    # glob_attrs = {**outdict["spin"].attrs["GLOBAL"], **var}
    glob_attrs = {
        "delta_energy_plus": e_plus.data,
        "delta_energy_minus": e_minus.data,
        "species": specie,
        **outdict["spin"].attrs["GLOBAL"],
    }

    # Build Dataset
    out = xr.Dataset(outdict, attrs=glob_attrs)

    return out



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# Built-in imports
from typing import Sequence

# 3rd party imports
from scipy import constants
from xarray.core.dataarray import DataArray

# Local imports
from .avg_4sc import avg_4sc
from .c_4_grad import c_4_grad
from .cross import cross

__author__ = "Louis Richard"
__email__ = "louisr@irfu.se"
__copyright__ = "Copyright 2020-2023"
__license__ = "MIT"
__version__ = "2.4.2"
__status__ = "Prototype"


def c_4_j(
    r_list: Sequence[DataArray], b_list: Sequence[DataArray]
) -> tuple[DataArray, DataArray, DataArray, DataArray, DataArray, DataArray]:
    r"""Calculate current density :math:`J` from using 4
    spacecraft technique [1]_, the divergence of the magnetic field
    :math:`\nabla . B`, magnetic field at the center of
    mass of the tetrahedron, :math:`J \times B`
    force, part of the divergence of stress associated with
    curvature :math:`\nabla.T_{shear}` and gradient of
    the magnetic pressure :math:`\nabla P_b`. Where :

    .. math::

        J = \frac{\nabla \times B}{\mu_0}

        J \times B = \nabla.T_{shear} + \nabla P_b

        \nabla.T_{shear} = \frac{(B.\nabla) B}{\mu_0}

        \nabla P_b = \nabla \frac{B^2}{2\mu_0}

    The divergence of the magnetic field is current density units as
    it shows the error on the estimation of the current density [2]_ .

    Parameters
    ----------
    r_list : list of DataArray
        Time series of the spacecraft position [km].
    b_list : list of DataArray
        Time series of the magnetic field [nT].

    Returns
    -------
    j : DataArray
        Time series of the current density [A.m^{-2}].
    div_b : DataArray
        Time series of the divergence of the magnetic field [A.m^{-2}].
    b_avg : DataArray
        Time series of the magnetic field at the center of mass of the
        tetrahedron, sampled at 1st SC time steps [nT].
    jxb : DataArray
        Time series of the :math:`J\timesB` force
        [T.A].
    div_t_shear : DataArray
        Time series of the part of the divergence of stress associated
        with curvature units [T A/m^2].
    div_pb : DataArray
        Time series of the gradient of the magnetic pressure.

    See also
    --------
    pyrfu.pyrf.c_4_k : Calculates reciprocal vectors in barycentric
                            coordinates.
    pyrfu.pyrf.c_4_grad : Calculate gradient of physical field using 4
                            spacecraft technique.

    References
    ----------
    .. [1]	Dunlop, M. W., A. Balogh, K.-H. Glassmeier, and P. Robert
            (2002a), Four-point Cluster application of magnetic field
            analysis tools: The Curl- ometer, J. Geophys. Res.,
            107(A11), 1384, doi : https://doi.org/10.1029/2001JA005088.

    .. [2]	Robert, P., et al. (1998), Accuracy of current determination,
            in Analysis Methods for Multi-Spacecraft Data, edited by G.
            Paschmann and P. W. Daly, pp. 395–418, Int. Space Sci. Inst.,
            Bern. url : http://www.issibern.ch/forads/sr-001-16.pdf

    Examples
    --------
    >>> import numpy as np
    >>> from pyrfu import mms, pyrf

    Time interval

    >>> tint = ["2019-09-14T07:54:00.000", "2019-09-14T08:11:00.000"]

    Spacecraft indices

    >>> mms_list = np.arange(1,5)

    Load magnetic field and spacecraft position

    >>> b_mms = [mms.get_data("b_gse_fgm_srvy_l2", tint, i) for i in mms_list]
    >>> r_mms = [mms.get_data("r_gse_mec_srvy_l2", tint, i) for i in mms_list]

    Compute current density, divergence of b, etc. using curlometer technique

    >>> j_xyz, _, b_xyz, _, _, _ = pyrf.c_4_j(r_mms, b_mms)

    """
    # Check inputs
    assert isinstance(r_list, list) and len(r_list) == 4, "r_list must a list of s/c"
    assert isinstance(b_list, list) and len(b_list) == 4, "b_list must a list of s/c"

    b_avg = 1e-9 * avg_4sc(b_list)

    # Estimate divB/mu0. unit is A/m2
    div_b = c_4_grad(r_list, b_list, "div")

    # to get right units
    div_b *= 1.0e-3 * 1e-9 / constants.mu_0

    # estimate current j [A/m2]
    j = c_4_grad(r_list, b_list, "curl")
    j.data *= 1.0e-3 * 1e-9 / constants.mu_0

    # estimate jxB force [T A/m2]
    jxb = cross(j, b_avg)

    # estimate divTshear = (1/muo) (B*div)B [T A/m2]
    div_t_shear = c_4_grad(r_list, b_list, "bdivb")
    div_t_shear.data *= 1.0e-3 * 1e-9 * 1e-9 / constants.mu_0

    # estimate divPb = (1/muo) grad (B^2/2) = divTshear-jxB
    div_pb = div_t_shear.copy()
    div_pb.data -= jxb.data

    return j, div_b, b_avg, jxb, div_t_shear, div_pb



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# 3rd party imports
import numpy as np

# Local imports
from pyrfu.pyrf.histogram2d import histogram2d
from pyrfu.pyrf.optimize_nbins_2d import optimize_nbins_2d

__author__ = "Louis Richard"
__email__ = "louisr@irfu.se"
__copyright__ = "Copyright 2020-2025"
__license__ = "MIT"
__version__ = "2.4.14"
__status__ = "Prototype"


def brazil(
    beta_para: np.ndarray,
    p_aniso: np.ndarray,
    bins: list = None,
    threshold: int = 9,
    **kwargs,
):
    """
    Computes 2D histogram and PDF (Brazil plot style) for plasma data.

    Parameters
    ----------
    beta_para : np.ndarray
        Parallel beta values (must be positive and finite).
    p_aniso : np.ndarray
        Temperature anisotropy values (must be positive and finite).
    bins : list, optional
        Bin edges or number of bins for the histogram. If None, optimized
        bin count is used.
    threshold : int, optional
        Minimum count threshold for masking low-counts (default is 9).

    Returns
    -------
    n : xarray.DataArray
        2D histogram counts with low-counts masked (NaN for <9).
    h : xarray.DataArray
        2D probability density function with low-counts masked.
    """
    # Valid data mask
    valid = (
        np.isfinite(beta_para) & (beta_para > 0) & np.isfinite(p_aniso) & (p_aniso > 0)
    )
    beta_para = beta_para[valid]
    p_aniso = p_aniso[valid]

    log_beta = np.log10(beta_para)
    log_aniso = np.log10(p_aniso)

    if bins is None:
        bins = optimize_nbins_2d(log_beta, log_aniso, **kwargs)

    # Compute histogram edges
    _, x_edges, y_edges = np.histogram2d(log_beta, log_aniso, bins=bins, density=True)

    # Compute bin centers in linear space
    x_centers = 10 ** (x_edges[:-1] + np.diff(x_edges) / 2)
    y_centers = 10 ** (y_edges[:-1] + np.diff(y_edges) / 2)

    # Histogram counts (not normalized)
    n = histogram2d(beta_para, p_aniso, bins=[10**x_edges, 10**y_edges], density=False)
    n = n.assign_coords({"x_bins": x_centers, "y_bins": y_centers})
    n.data[n.data == 0] = np.nan  # mask zero counts

    # Histogram PDF (normalized)
    h = histogram2d(beta_para, p_aniso, bins=[10**x_edges, 10**y_edges], density=True)
    h = h.assign_coords({"x_bins": x_centers, "y_bins": y_centers})
    h.data[n.data < threshold] = np.nan  # apply count threshold mask to PDF

    return n, h



#!/usr/bin/env python
# -*- coding: utf-8 -*-

from typing import Literal, Optional, Union

# 3rd party imports
import numpy as np
import xarray as xr
from xarray.core.dataarray import DataArray
from xarray.core.dataset import Dataset

# Local imports
from ..pyrf.calc_fs import calc_fs
from ..pyrf.cotrans import cotrans
from ..pyrf.resample import resample
from ..pyrf.ts_vec_xyz import ts_vec_xyz

__author__ = "Louis Richard"
__email__ = "louisr@irfu.se"
__copyright__ = "Copyright 2020-2023"
__license__ = "MIT"
__version__ = "2.4.2"
__status__ = "Prototype"


def _transformation_matrix(
    spin_axis: np.ndarray, direction: Literal[-1, 1]
) -> np.ndarray:
    r"""Return transformation matrix from DSL to GSE.

    Parameters
    ----------
    spin_axis : numpy.ndarray
        Spin axis.
    direction : {-1, 1}
        Direction of the transformation.

    Returns
    -------
    numpy.ndarray
        Transformation matrix.

    Raises
    ------
    ValueError
        If direction is not -1 or 1.

    """
    r_x, r_y, r_z = [spin_axis[:, i] for i in range(3)]

    fact = 1.0 / np.sqrt(r_y**2 + r_z**2)
    out = np.zeros((len(fact), 3, 3))
    out[:, 0, :] = np.transpose(
        np.stack(
            [
                fact * (r_y**2 + r_z**2),
                -fact * r_x * r_y,
                -fact * r_x * r_z,
            ],
        ),
    )

    out[:, 1, :] = np.transpose(
        np.stack([0.0 * fact, fact * r_z, -fact * r_y]),
    )

    out[:, 2, :] = np.transpose(np.stack([r_x, r_y, r_z]))

    if direction == 1:
        out = np.transpose(out, [0, 2, 1])
    elif direction == -1:
        out = np.transpose(out, [0, 1, 2])
    else:
        raise ValueError("Direction must be either 1 or -1.")

    return out


def dsl2gse(
    inp: DataArray, defatt: Union[Dataset, np.ndarray], direction: Optional[int] = 1
):
    r"""Transform time series from MMS's DSL to GSE.

    Parameters
    ----------
    inp : DataArray
        Input time series to convert.
    defatt : Dataset or numpy.ndarray
        Spacecraft attitude.
    direction : {1, -1}, Optional
        Direction of transformation. +1 DSL -> GSE, -1 GSE -> DSL.
        Default is 1.

    Returns
    -------
    out : xarray.DataArray
        Time series of the input field in the new coordinates systems.

    Raises
    ------
    TypeError
        If defatt is not a Dataset or a numpy.ndarray.

    Examples
    --------
    >>> from pyrfu.mms import get_data, load_ancillary, dsl2gse

    Define time interval

    >>> tint = ["2015-05-09T14:00:000", "2015-05-09T17:59:590"]

    Load magnetic field in spacecraft coordinates

    >>> b_xyz = get_data("b_dmpa_fgm_brst_l2", tint, 1)

    Load spacecraft attitude

    >>> defatt = load_ancillary("defatt", tint, 1)

    Transform magnetic field to GSE

    >>> b_gse = dsl2gse(b_xyz, defatt)

    """
    if isinstance(defatt, xr.Dataset):
        x = np.cos(np.deg2rad(defatt.z_dec)) * np.cos(
            np.deg2rad(defatt.z_ra.data),
        )
        y = np.cos(np.deg2rad(defatt.z_dec)) * np.sin(
            np.deg2rad(defatt.z_ra.data),
        )
        z = np.sin(np.deg2rad(defatt.z_dec))
        sax_gei = ts_vec_xyz(defatt.time.data, np.transpose(np.vstack([x, y, z])))
        sax_gse = cotrans(sax_gei, "gei>gse")
        spin_ax_gse = resample(sax_gse, inp, f_s=calc_fs(inp))
        spin_axis = spin_ax_gse.data

    elif isinstance(defatt, np.ndarray) and len(defatt) == 3:
        spin_axis = np.atleast_2d(defatt)

    else:
        raise TypeError("DEFATT/SAX input must be xarray.Dataset or vector")

    # Compute transformation natrix
    transf_mat = _transformation_matrix(spin_axis, direction)

    out_data = np.einsum("kji,ki->kj", transf_mat, inp.data)

    out = inp.copy()
    out.data = out_data
    out.attrs["COORDINATE_SYSTEM"] = "GSE"

    return out



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# Built-in imports
import itertools

# 3rd party imports
import numpy as np
import xarray as xr

# Local imports
from .dsl2gse import dsl2gse

__author__ = "Louis Richard"
__email__ = "louisr@irfu.se"
__copyright__ = "Copyright 2020-2023"
__license__ = "MIT"
__version__ = "2.4.2"
__status__ = "Prototype"


def _check_spin(spin):
    _, spin_inds = np.unique(spin, return_index=True)

    len_spin = np.max(np.diff(spin_inds))

    # Check to see if we have a complete first spin,
    # if not then go to the second spin
    if np.where(spin == spin[spin_inds[0]])[0].size != len_spin:
        spin_inds = spin_inds[1:]

    # Check to see if the last one is complete,
    # if not then go to the second spin
    if np.where(spin == spin[spin_inds[-1]])[0].size != len_spin:
        spin_inds = spin_inds[:-1]

    return spin_inds, len_spin


def _combine_attrs(inp_allt_attrs):
    attrs = {}
    for k in inp_allt_attrs[0]:
        allt_attrs = [inp_allt_attr[k] for inp_allt_attr in inp_allt_attrs]
        is_same = [allt_attr == allt_attrs[0] for allt_attr in allt_attrs[1:]]

        if all(is_same):
            attrs[k] = allt_attrs[0]
        else:
            continue

    return attrs


def eis_ang_ang(inp_allt, en_chan: list = None, defatt: xr.Dataset = None):
    r"""Generates EIS angle-angle distribution.

    Parameters
    ----------
    inp_allt : xarray.Dataset
        Dataset of the fluxes of all 6 telescopes.
    en_chan : array_like, Optional
        Energy channels to use. Default use all energy channels.

    Returns
    -------
    out : xarray.DataArray
        EIS skymap-like distribution.

    """

    if en_chan is None:
        en_chan = np.arange(len(inp_allt.energy))

    time_ = inp_allt.time.data

    n_en = len(en_chan)

    scopes = list(filter(lambda x: x.startswith("t"), inp_allt.keys()))

    phi, theta = [np.zeros((len(scopes), len(inp_allt.time))) for _ in range(2)]

    for i, scope in enumerate(scopes):
        d_xyz = inp_allt[f"look_{scope}"]

        # If defatt is given get angles in spacecraft coordinates system
        if defatt is not None:
            d_xyz = dsl2gse(d_xyz, defatt, -1)
            coordinates_system = "DBCS>Despun Body Coordinate System"
        else:
            coordinates_system = "GSE>Geocentric Solar Magnetospheric"

        # Domain [-180, 180], 0 = sunward (GSE)
        # phi[i, :] = (np.rad2deg(np.arctan2(d_xyz.data[:, 1], d_xyz.data[:, 0])))
        # Domain [0, 360], 0 = sunward (GSE)
        phi[i, :] = np.rad2deg(np.arctan2(d_xyz.data[:, 1], d_xyz.data[:, 0])) + 180.0
        # Domain [-90, 90], Positive is look direction northward
        # theta[i, :] = 90.0 - np.rad2deg(np.arccos(d_xyz[:, 2]))
        # Domain [0, 180], Positive is look direction northward
        theta[i, :] = np.rad2deg(np.arccos(d_xyz[:, 2]))

    spin_ = inp_allt.spin.data
    sect_ = inp_allt.sector.data

    spin_inds, len_spin = _check_spin(spin_)

    n_spins, n_pol, n_azi = [len(spin_inds), 6, len(np.unique(sect_))]

    # Minus 80 plus
    min_pol_edges = -80.0 + 160.0 * np.arange(n_pol) / n_pol
    max_pol_edges = -80.0 + 160.0 * (np.arange(n_pol) + 1) / n_pol
    mid_pol_edges = min_pol_edges + 90.0 / n_pol

    min_azi_edges = -180.0 + 360.0 * np.arange(n_azi) / n_azi
    max_azi_edges = -180.0 + 360.0 * (np.arange(n_azi) + 1) / n_azi
    mid_azi_edges = min_azi_edges + 180.0 / n_azi

    out_data = np.zeros((n_spins, n_en, n_azi, n_pol))

    time_data = time_[spin_inds + len_spin // 2]

    for i, spin_ind in enumerate(spin_inds):
        t_inds = np.where(spin_ == spin_[spin_ind])[0]
        for t_ind, (i_s, scope) in itertools.product(
            t_inds,
            enumerate(scopes),
        ):
            cond_azi = np.logical_and(
                phi[i_s, t_ind] > min_azi_edges,
                phi[i_s, t_ind] < max_azi_edges,
            )
            a_idx = np.where(cond_azi)[0]

            cond_pol = np.logical_and(
                theta[i_s, t_ind] > min_pol_edges,
                theta[i_s, t_ind] < max_pol_edges,
            )
            p_idx = np.where(cond_pol)[0]

            out_data[i, :, a_idx, p_idx] = inp_allt[scope].data[t_ind, en_chan]

    # Setup attributes (that of all telescopes + delta energies and
    # particle species)
    attrs = {k: inp_allt.attrs[k] for k in ["delta_energy_plus", "delta_energy_minus"]}
    attrs = {"species": inp_allt.attrs["species"], **attrs}
    attrs = {
        **attrs,
        **_combine_attrs([inp_allt[scope].attrs for scope in scopes]),
    }
    attrs = {k: attrs[k] for k in sorted(attrs)}

    out = xr.DataArray(
        out_data,
        coords=[
            time_data,
            inp_allt.energy.data[en_chan],
            mid_azi_edges + 180.0,
            mid_pol_edges + 90.0,
        ],
        dims=["time", "energy", "phi", "theta"],
        attrs=attrs,
    )

    # Fill coordinates attributes
    out.time.attrs = inp_allt.time.attrs  # time
    out.energy.attrs = inp_allt.energy.attrs  # energy

    # Fill azimuthal angles attributes
    out.phi.attrs = {
        "CATDESC": f"{attrs['CATDESC'][0][:4]} EPD-EIS sky-map azimuth angles",
        "COORDINATE_SYSTEM": coordinates_system,
        "DELTA_MINUS_VAR": " ",
        "DELTA_PLUS_VAR": " ",
        "FIELDNAM": f"{attrs['CATDESC'][0][:4]} EPD-EIS phi",
        "FILLVAL": -1e31,
        "FORMAT": "E12.2",
        "LABLAXIS": "phi",
        "REPRESENTATION_1": "t",
        "SCALETYP": "linear",
        "SI_CONVERSION": "0.0174532925>rad",
        "UNITS": "deg",
        "VALIDMAX": 360.0,
        "VALIDMIN": 0.0,
        "VAR_NOTES": "Azimuth angles in the instrument frame",
        "VAR_TYPE": "support_data",
    }

    # Fill elevation angles attributes
    out.theta.attrs = {
        "CATDESC": f"{attrs['CATDESC'][0][:4]} EPD-EIS sky-map zenith angles",
        "COORDINATE_SYSTEM": coordinates_system,
        "DELTA_MINUS_VAR": " ",
        "DELTA_PLUS_VAR": " ",
        "FIELDNAM": f"{attrs['CATDESC'][0][:4]} EPD-EIS theta",
        "FILLVAL": -1e31,
        "FORMAT": "E12.2",
        "LABLAXIS": "theta",
        "REPRESENTATION_1": "t",
        "SCALETYP": "linear",
        "SI_CONVERSION": "0.0174532925>rad",
        "UNITS": "deg",
        "VALIDMAX": 180.0,
        "VALIDMIN": 0.0,
        "VAR_NOTES": "Pixel zenith angles",
        "VAR_TYPE": "support_data",
    }

    return out



#!/usr/bin/env python
# -*- coding: utf-8 -*-

from .anisotropy_thresholds import anisotropy_thresholds
from .autocorr import autocorr
from .average_vdf import average_vdf
from .avg_4sc import avg_4sc
from .brazil import brazil
from .c_4_grad import c_4_grad
from .c_4_j import c_4_j
from .c_4_k import c_4_k
from .c_4_v import c_4_v
from .calc_ag import calc_ag
from .calc_agyro import calc_agyro
from .calc_dng import calc_dng
from .calc_dt import calc_dt
from .calc_fs import calc_fs
from .calc_sqrtq import calc_sqrtq
from .cart2sph import cart2sph
from .cart2sph_ts import cart2sph_ts
from .cdfepoch2datetime64 import cdfepoch2datetime64
from .compress_cwt import compress_cwt
from .convert_fac import convert_fac
from .corr_deriv import corr_deriv
from .cotrans import cotrans
from .cross import cross
from .date_str import date_str
from .datetime2iso8601 import datetime2iso8601
from .datetime642iso8601 import datetime642iso8601
from .datetime642ttns import datetime642ttns
from .datetime642unix import datetime642unix
from .dec_par_perp import dec_par_perp
from .dist_append import dist_append
from .dot import dot
from .dynamic_press import dynamic_press
from .e_vxb import e_vxb
from .eb_nrf import eb_nrf
from .ebsp import ebsp
from .edb import edb
from .end import end
from .estimate import estimate
from .extend_tint import extend_tint
from .filt import filt
from .find_closest import find_closest
from .get_omni_data import get_omni_data
from .gradient import gradient
from .gse2gsm import gse2gsm
from .histogram import histogram
from .histogram2d import histogram2d
from .increments import increments
from .int_sph_dist import int_sph_dist
from .integrate import integrate
from .iplasma_calc import iplasma_calc
from .iso86012datetime import iso86012datetime
from .iso86012datetime64 import iso86012datetime64
from .iso86012timevec import iso86012timevec
from .iso86012unix import iso86012unix
from .l_shell import l_shell
from .lowpass import lowpass
from .magnetosphere import magnetosphere
from .match_phibe_dir import match_phibe_dir
from .match_phibe_v import match_phibe_v
from .mean import mean
from .mean_bins import mean_bins
from .mean_field import mean_field
from .medfilt import medfilt
from .median_bins import median_bins
from .movmean import movmean
from .mva import mva
from .mva_gui import mva_gui
from .nanavg_4sc import nanavg_4sc
from .new_xyz import new_xyz
from .norm import norm
from .normalize import normalize
from .optimize_nbins_1d import optimize_nbins_1d
from .optimize_nbins_2d import optimize_nbins_2d
from .pid_4sc import pid_4sc
from .plasma_beta import plasma_beta
from .plasma_calc import plasma_calc
from .poynting_flux import poynting_flux
from .pres_anis import pres_anis
from .psd import psd
from .pvi import pvi
from .pvi_4sc import pvi_4sc
from .read_cdf import read_cdf
from .remove_repeated_points import remove_repeated_points
from .resample import resample
from .shock_normal import shock_normal
from .shock_parameters import shock_parameters
from .sliding_derivative import sliding_derivative
from .solid_angle import solid_angle
from .sph2cart import sph2cart
from .st_diff import st_diff
from .start import start
from .struct_func import struct_func
from .t_eval import t_eval
from .time_clip import time_clip
from .timevec2iso8601 import timevec2iso8601
from .trace import trace
from .ts_append import ts_append
from .ts_convolve import ts_convolve
from .ts_scalar import ts_scalar
from .ts_skymap import ts_skymap
from .ts_spectr import ts_spectr
from .ts_tensor_xyz import ts_tensor_xyz
from .ts_time import ts_time
from .ts_vec_xyz import ts_vec_xyz
from .ttns2datetime64 import ttns2datetime64
from .unix2datetime64 import unix2datetime64
from .vht import vht
from .wave_fft import wave_fft
from .wavelet import wavelet
from .wavepolarize_means import wavepolarize_means
from .waverage import waverage

__author__ = "Louis Richard"
__email__ = "louisr@irfu.se"
__copyright__ = "Copyright 2020-2024"
__license__ = "MIT"
__version__ = "2.4.14"
__status__ = "Prototype"

__all__ = [
    "anisotropy_thresholds",
    "autocorr",
    "average_vdf",
    "avg_4sc",
    "brazil",
    "c_4_grad",
    "c_4_j",
    "c_4_k",
    "c_4_v",
    "calc_ag",
    "calc_agyro",
    "calc_dng",
    "calc_dt",
    "calc_fs",
    "calc_sqrtq",
    "cart2sph",
    "cart2sph_ts",
    "cdfepoch2datetime64",
    "compress_cwt",
    "convert_fac",
    "corr_deriv",
    "cotrans",
    "cross",
    "date_str",
    "datetime2iso8601",
    "datetime642iso8601",
    "datetime642ttns",
    "datetime642unix",
    "dec_par_perp",
    "dist_append",
    "dot",
    "dynamic_press",
    "e_vxb",
    "eb_nrf",
    "ebsp",
    "edb",
    "end",
    "estimate",
    "extend_tint",
    "filt",
    "find_closest",
    "get_omni_data",
    "gradient",
    "gse2gsm",
    "histogram",
    "histogram2d",
    "increments",
    "int_sph_dist",
    "integrate",
    "iplasma_calc",
    "iso86012unix",
    "iso86012datetime",
    "iso86012datetime64",
    "iso86012timevec",
    "l_shell",
    "lowpass",
    "magnetosphere",
    "match_phibe_dir",
    "match_phibe_v",
    "mean",
    "mean_bins",
    "mean_field",
    "medfilt",
    "median_bins",
    "movmean",
    "mva",
    "mva_gui",
    "nanavg_4sc",
    "new_xyz",
    "norm",
    "normalize",
    "optimize_nbins_1d",
    "optimize_nbins_2d",
    "pid_4sc",
    "plasma_beta",
    "plasma_calc",
    "poynting_flux",
    "pres_anis",
    "psd",
    "pvi",
    "pvi_4sc",
    "read_cdf",
    "remove_repeated_points",
    "resample",
    "shock_normal",
    "shock_parameters",
    "sliding_derivative",
    "solid_angle",
    "sph2cart",
    "st_diff",
    "start",
    "struct_func",
    "t_eval",
    "time_clip",
    "timevec2iso8601",
    "trace",
    "ts_append",
    "ts_convolve",
    "ts_scalar",
    "ts_spectr",
    "ts_skymap",
    "ts_tensor_xyz",
    "ts_time",
    "ts_vec_xyz",
    "ttns2datetime64",
    "unix2datetime64",
    "vht",
    "wave_fft",
    "wavelet",
    "wavepolarize_means",
    "waverage",
]



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# 3rd party imports
import numpy as np
import xarray as xr
from scipy import constants, special

# Local imports
from ..pyrf.resample import resample

__author__ = "Louis Richard"
__email__ = "louisr@irfu.se"
__copyright__ = "Copyright 2020-2023"
__license__ = "MIT"
__version__ = "2.4.2"
__status__ = "Prototype"


def _thermal_velocity(energy, specie):
    if specie.lower() == "ions":
        mass = constants.proton_mass
    elif specie.lower() == "electrons":
        mass = constants.electron_mass
    else:
        raise ValueError("Invalid specie")

    return np.sqrt(2 * energy * constants.electron_volt / mass)


def make_model_kappa(vdf, n_s, v_xyz_s, t_s, kappa: float = 7.0):
    r"""Make a general isottropic kappa distribution function based on
    particle moment data in the same format as vdf.

    Parameters
    ----------
    vdf : xarray.Dataset
        Particle distribution (skymap).
    n_s : xarray.DataArray
        Time series of the number density.
    v_xyz_s : xarray.DataArray
        Time series of the bulk velocity.
    t_s : xarray.DataArray
        Time series of the scalar temperature.
    kappa : float. Optional
        Kappa index, Default is 7.

    Returns
    -------
    out : xarray.Dataset
        Distribution function in the same format as vdf.

    See also
    --------
    pyrfu.mms.make_model_vdf

    Todo
    ----
    Generalize to bi-Kappa distributions.

    """

    # Unpack azimuthal and elevation angles from the skymap distribution

    phi = vdf.phi.data[0, :]
    theta = vdf.theta.data

    # Resample number density, bulk velocity and temperature to skymap sampling
    n_s = resample(n_s, vdf.time)
    v_xyz_s = resample(v_xyz_s, vdf.time)
    t_s = resample(t_s, vdf.time)

    # Compute thermal velocity
    v_th = _thermal_velocity(t_s.data, vdf.attrs["species"])

    # Initialize output to zeros
    out_data = np.zeros(
        [
            vdf.data.shape[0],
            vdf.data.shape[3],
            vdf.data.shape[1],
            vdf.data.shape[2],
        ],
    )

    for i in range(len(vdf.time.data)):
        energies = vdf.energy.data[i, :]

        # Construct velocities matrix
        en_mat, theta_mat, phi_mat = np.meshgrid(energies, theta, phi)
        v_x_hat = np.sin(np.deg2rad(theta_mat)) * np.cos(np.deg2rad(phi_mat))
        v_y_hat = np.sin(np.deg2rad(theta_mat)) * np.sin(np.deg2rad(phi_mat))
        v_z_hat = np.cos(np.deg2rad(theta_mat))
        v_mag = _thermal_velocity(en_mat, vdf.attrs["species"])
        v_mat = np.stack([v_mag * v_x_hat, v_mag * v_y_hat, v_mag * v_z_hat])

        # Unpack thermal and bulk velocities
        v_thk = (kappa - 3 / 2) * v_th[i] ** 2
        v_bulk = v_xyz_s.data[i, ...]

        # Shift velocities with lbulk velocity
        v_squ = np.sum((v_mat - v_bulk[:, None, None, None]) ** 2, axis=0)

        # Compute kappa distribution
        coef1 = 1 / (np.pi * v_thk) ** (3 / 2)
        coef2 = special.gamma(kappa + 1) / special.gamma(kappa - 1 / 2)
        out_data[i, ...] = (1 + v_squ / v_thk) ** -(kappa + 1)
        out_data[i, ...] *= coef1 * coef2

    # Scale with number density
    out_data *= n_s.data[:, None, None, None]

    # Convert to differential particule flux
    # out_data *= np.tile(en_mat, (n_t, 1, 1, 1)) * 1e15 / 0.53707

    # Transform to usual shape
    out_data = np.transpose(out_data, [0, 2, 3, 1])

    # Creates Dataset
    out_dict = {f"idx{i}": range(k) for i, k in enumerate(out_data.shape[1:])}
    out_dict["data"] = (["time", "idx0", "idx1", "idx2"], out_data)
    out_dict["energy"] = (["time", "idx0"], vdf.energy.data)
    out_dict["phi"] = (["time", "idx1"], vdf.phi.data)
    out_dict["theta"] = (["idx2"], vdf.theta.data)
    out_dict["time"] = vdf.time.data

    out = xr.Dataset(out_dict)

    out.attrs["UNITS"] = "s^3/m^6"
    out.attrs["species"] = vdf.attrs.get("species", None)
    out.attrs["energy_dminus"] = vdf.attrs.get("energy_dminus", None)
    out.attrs["energy_dplus"] = vdf.attrs.get("energy_dplus", None)

    return out



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# 3rd party imports
import numpy as np
import xarray as xr

# Local imports
from .read_feeps_sector_masks_csv import read_feeps_sector_masks_csv

__author__ = "Louis Richard"
__email__ = "louisr@irfu.se"
__copyright__ = "Copyright 2020-2023"
__license__ = "MIT"
__version__ = "2.4.2"
__status__ = "Prototype"


def feeps_remove_sun(inp_dataset):
    r"""Removes the sunlight contamination from FEEPS data.

    Parameters
    ----------
    inp_dataset : xarray.Dataset
        Dataset of energy spectrum of all eyes.

    Returns
    -------
    out : xarray.Dataset
        Dataset of cleaned energy spectrum of all eyes.

    See also
    --------
    pyrfu.mms.get_feeps_alleyes : Read energy spectrum for all FEEPS eyes.

    Examples
    --------
    >>> from pyrfu import mms

    Define time interval

    >>> tint = ["2017-07-18T13:04:00.000", "2017-07-18T13:07:00.000"]

    Spacecraft index

    >>> mms_id = 2

    Load data from FEEPS

    >>> cps_i = mms.get_feeps_alleyes("CPSi_brst_l2", tint, mms_id)
    >>> cps_i_clean, _ = mms.feeps_split_integral_ch(cps_i)
    >>> cps_i_clean_sun_removed = mms.feeps_remove_sun(cps_i_clean)

    """

    var = inp_dataset.attrs

    tint = list(np.datetime_as_string(inp_dataset.time.data[[0, -1]], "ns"))

    spin_sectors = inp_dataset["spinsectnum"]
    mask_sectors = read_feeps_sector_masks_csv(tint)

    out_dict = {}
    out_dict["spinsectnum"] = inp_dataset["spinsectnum"]

    for k in inp_dataset:
        out_dict[k] = inp_dataset[k]
        if mask_sectors.get(f"mms{var['mmsId']:d}_imask_{k}") is not None:
            bad_sectors = mask_sectors[f"mms{var['mmsId']:d}_imask_{k}"]

            for bad_sector in bad_sectors:
                this_bad_sector = np.where(spin_sectors == bad_sector)[0]
                if len(this_bad_sector) != 0:
                    out_dict[k].data[this_bad_sector] = np.nan

    out = xr.Dataset(out_dict, attrs=var)

    out.attrs = var

    return out



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# Built-in imports
import json
import logging
import os
from typing import Literal, Optional

# 3rd party imports
import keyring
from keyrings.alt.file import PlaintextKeyring

__author__ = "Louis Richard"
__email__ = "louisr@irfu.se"
__copyright__ = "Copyright 2020-2024"
__license__ = "MIT"
__version__ = "2.4.13"
__status__ = "Prototype"

logging.captureWarnings(True)
logging.basicConfig(
    format="[%(asctime)s] %(levelname)s: %(message)s",
    datefmt="%d-%b-%y %H:%M:%S",
    level=logging.INFO,
)

MMS_CFG_PATH = os.path.join(os.path.dirname(os.path.abspath(__file__)), "config.json")


def db_init(
    default: Optional[Literal["local", "sdc", "aws"]] = "local",
    local: Optional[str] = ".",
    sdc: Optional[str] = "public",
    sdc_username: Optional[str] = "username",
    sdc_password: Optional[str] = "password",
    aws: Optional[str] = "",
) -> None:
    r"""Manage the MMS data access configuration.

    The default resource to access MMS data, the local path to use, the MMS SDC
    credentials saved in encrypted file in your home directory, and the Amazon Web
    Services (AWS) bucket name.

    Parameters
    ----------
    default : {"local", "sdc", "aws"}, Optional
        Name of the default resource to access the MMS data. Default is local.
    local : str, Optional
        Local path to MMS data. Default is /Volumes/mms.
    sdc : {"public", "sitl"}, Optional
        Rights to access MMS data from SDC. If "sitl" please make sure to register
        valid SDC credential. Default is public.
    sdc_username : str, Optional
        MMS SDC credential username. Default is "username".
    sdc_password : str, Optional
        MMS SDC credential password. Default is "password".
    aws : str, Optional
        Bucket name and prefix to MMS data in AWS S3. Default is empty.

    Raises
    ------
    NotImplementedError
        If the default resource is not implemented.
    FileNotFoundError
        If the local path doesn't exist.
    ValueError
        If the SDC rights are not "public" or "sitl".

    """
    keyring.set_keyring(PlaintextKeyring())

    # Check default
    if default.lower() not in ["local", "sdc", "aws"]:
        raise NotImplementedError(f"Resource {default} is not implemented!!")

    # Normalize the path and make sure that it exists
    local = os.path.normpath(os.path.abspath(local))

    if not os.path.exists(local):
        raise FileNotFoundError(f"{local} doesn't exists!!")

    # Check MMS SDC rights
    if sdc.lower() not in ["public", "sitl"]:
        raise ValueError("sdc must be 'public' or 'sitl'!!")

    config = {
        "default": default.lower(),
        "local": local,
        "sdc": {"rights": sdc.lower(), "username": sdc_username},
        "aws": aws,
    }

    logging.info("Updating MMS data access configuration in %s...", MMS_CFG_PATH)

    # Overwrite the configuration file with the new path
    with open(MMS_CFG_PATH, "w", encoding="utf-8") as fs:
        json.dump(config, fs)

    # Read credentials for sdc_username
    credential_path = str(keyring.util.platform_.config_root())
    credential = keyring.get_credential("mms-sdc", sdc_username)

    if (
        not credential
        or credential.username == "username"
        or credential.password == "password"
    ):
        # if credentials are empty overwrite anyway
        username, password = sdc_username, sdc_password
    elif sdc_username == "username" or sdc_password == "password":
        # if existing credentials and incomplete arguments do not overwrite
        username, password = credential.username, credential.password
    else:
        # if existing credentials and complete arguments overwrite
        username, password = sdc_username, sdc_password

    logging.info("Updating MMS SDC credentials in %s...", credential_path)

    keyring.set_password("mms-sdc", username, password)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# Built-in imports
import json
import logging
import os
import warnings
from datetime import datetime, timedelta
from shutil import copy, copyfileobj
from tempfile import NamedTemporaryFile

# 3rd party imports
import numpy as np
import requests
import tqdm

__author__ = "Louis Richard"
__email__ = "louisr@irfu.se"
__copyright__ = "Copyright 2020-2023"
__license__ = "MIT"
__version__ = "2.4.10"
__status__ = "Prototype"

logging.captureWarnings(True)
logging.basicConfig(
    format="[%(asctime)s] %(levelname)s: %(message)s",
    datefmt="%d-%b-%y %H:%M:%S",
    level=logging.INFO,
)


LASP_PUBL = "https://lasp.colorado.edu/maven/sdc/public/files/api/v1/"


def _login_lasp(user: str, password: str):
    r"""Login to LASP colorado."""

    session = requests.Session()
    session.auth = (user, password)

    with warnings.catch_warnings():
        warnings.simplefilter("ignore", category=ResourceWarning)
        _ = session.post("https://lasp.colorado.edu", verify=True, timeout=5)

    return session


def _construct_url(tint, var, lasp_url):
    r"""Construct the url that return a json-formatted string of science
    filenames that are available for download according to:
    https://lasp.colorado.edu/mms/sdc/team/about/how-to/
    """

    tint = np.array(tint).astype("<M8[ns]").astype(str)
    tint = [datetime.strptime(t_[:-3], "%Y-%m-%dT%H:%M:%S.%f") for t_ in tint]
    start_date = tint[0].strftime("%Y-%m-%d")
    end_date = tint[1].strftime("%Y-%m-%d")

    # If start and end are on the same add one day to end date
    if end_date == start_date:
        tint[1] += timedelta(days=1)
        end_date = tint[1].strftime("%Y-%m-%d")

    url = f"{lasp_url}search/science/fn_metadata/file_info?"

    url = f"{url}instrument={var['inst']}"
    url = f"{url}&level={var['level']}"
    url = f"{url}&start_date={start_date}&end_date={end_date}"

    if "file_extension" in var:
        url = f"{url}&file_extension={var['file_extension']}"

    # if "plan" in var:
    #     url = f"{url}&plan={var['plan']}"

    return url


def _make_path(file, var, lasp_url, data_path: str = ""):
    r"""Construct path of the data file using the standard convention."""

    if not data_path:
        pkg_path = os.path.dirname(os.path.abspath(__file__))

        # Read the current version of the MAVEN configuration file
        with open(os.path.join(pkg_path, "config.json"), "r", encoding="utf-8") as fs:
            config = json.load(fs)

        data_path = os.path.normpath(config["local_data_dir"])
    else:
        data_path = os.path.normpath(data_path)

    assert os.path.exists(data_path), "local data directory doesn't exist!"

    path_list = [
        data_path,  # root path to data
        var["inst"],  # instrument sub-directory
        var["level"],  # data level sub-directory
        file["file_name"].split("_")[4][:4],  # year sub-directory
        file["file_name"].split("_")[4][4:6],  # month sub-directory
    ]

    out_path = os.path.join(*path_list)
    out_file = os.path.join(*path_list, file["file_name"])

    download_url = (
        f"{lasp_url}search/science/fn_metadata/download?file={file['file_name']}"
    )

    return out_path, out_file, download_url


def download_data(var, tint, login: str = "", password: str = "", data_path: str = ""):
    r"""Downloads files containing field `var_str` over the time interval
    `tint`. The files are saved to `data_path`.

    Parameters
    ----------
    var : dict
        Hashtable containing:
            - inst: instrument acronym (acc, euv, iuv, kp, lpw, mag, ngi, pfp, sep, sta,
             swe, swi)
            - level: data levels (l1a, l1b, l2, l3, insitu (KP only), iuv (KP only))
    tint : list of str
        Time interval.
    login : str, Optional
        Login to LASP MAVEN. Default downloads from
        https://lasp.colorado.edu/maven/sdc/public/
    password : str, Optional
        Password to LASP MAVEN. Default downloads from
        https://lasp.colorado.edu/maven/sdc/public/
    data_path : str, Optional
        Path of MAVEN data. If None use `pyrfu/mms/config.json`

    """

    if not login:
        lasp_url = LASP_PUBL
    else:
        lasp_url = LASP_PUBL
        logging.info("login not impleted. Use public instead")

    sdc_session = _login_lasp(login, password)

    headers = {"User-Agent": "pyrfu"}

    url = _construct_url(tint, var, lasp_url)

    with warnings.catch_warnings():
        warnings.simplefilter("ignore", category=ResourceWarning)
        http_json = sdc_session.get(url, verify=True, headers=headers).json()

    if not http_json["files"]:
        raise FileNotFoundError("No files to download!!")

    for file in http_json["files"]:
        out_path, out_file, dwl_url = _make_path(file, var, lasp_url, data_path)
        plan = out_file.split("/")[-1].split("_")[3][7:]
        if plan == var["plan"] and out_file[-3:] == "sts":

            logging.info(
                "Downloading %s from %s...", os.path.basename(out_file), dwl_url
            )

            with warnings.catch_warnings():
                warnings.simplefilter("ignore", category=ResourceWarning)
                with sdc_session.get(
                    dwl_url,
                    stream=True,
                    verify=True,
                    headers=headers,
                ) as fsrc:
                    with NamedTemporaryFile(delete=False) as ftmp:
                        with tqdm.tqdm.wrapattr(
                            fsrc.raw,
                            "read",
                            total=file["file_size"],
                            ncols=60,
                        ) as fsrc_raw:
                            with open(ftmp.name, "wb") as fs:
                                copyfileobj(fsrc_raw, fs)

                    os.makedirs(out_path, exist_ok=True)

                    # if the download was successful, copy to data directory
                    copy(ftmp.name, out_file)

    sdc_session.close()



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# 3rd party imports
import matplotlib as mpl
import matplotlib.dates as mdates
import matplotlib.pyplot as plt
import matplotlib.ticker as mticker
import numpy as np
import xarray as xr

__author__ = "Louis Richard"
__email__ = "louisr@irfu.se"
__copyright__ = "Copyright 2020-2023"
__license__ = "MIT"
__version__ = "2.4.2"
__status__ = "Prototype"


def plot_line(axis, inp, **kwargs):
    r"""Line plot of time series.

    Parameters
    ----------
    axis : matplotlib.axes._axes.Axes
        Single axis where to plot inp. If None creates a new figure with a single axis.
    inp : xarray.DataArray
        Time series to plot

    Other Parameters
    ----------------
    **kwargs
        Keyword arguments control the line properties. See matplotlib.lines.Line2D
        for reference.

    Returns
    -------
    axs : matplotlib.axes._axes.Axes
        Axis with matplotlib.lines.Line2D.

    """

    if axis is None:
        _, axis = plt.subplots(1)
    else:
        if not isinstance(axis, mpl.axes.Axes):
            raise TypeError("axis must be a matplotlib.axes._axes.Axes")

    if not isinstance(inp, xr.DataArray):
        raise TypeError("inp must be an xarray.DataArray object!")

    if inp.data.ndim < 3:
        data = inp.data
    elif inp.data.ndim == 3:
        data = np.reshape(
            inp.data,
            (inp.shape[0], inp.shape[1] * inp.shape[2]),
        )
    else:
        raise NotImplementedError(
            f"plot_line cannot handle {inp.data.ndim} dimensional data"
        )

    time = inp.time
    axis.plot(time, data, **kwargs)

    if time.dtype == "<M8[ns]":
        locator = mdates.AutoDateLocator(minticks=3, maxticks=7)
        formatter = mdates.ConciseDateFormatter(locator)
        axis.xaxis.set_major_locator(locator)
        axis.xaxis.set_major_formatter(formatter)

    axis.grid(True, which="major", linestyle="-", linewidth="0.2", c="0.5")
    axis.yaxis.set_major_locator(mticker.MaxNLocator(4))

    return axis



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# 3rd party imports
from copy import deepcopy
from typing import Sequence

import numpy as np
import xarray as xr
from xarray.core.dataarray import DataArray

# Local imports
from ..pyrf.calc_fs import calc_fs
from ..pyrf.resample import resample

__author__ = "Apostolos Kolokotronis"
__email__ = "apostolos.kolokotronis@irf.se"
__copyright__ = "Copyright 2020-2024"
__license__ = "MIT"
__version__ = "2.4.2"
__status__ = "Prototype"


def _nan_count(inp):
    r"""Counts the number of non-NaN values in the input array at each time step.

    Parameters
    ----------
    inp : DataArray or Dataset
        Input array.

    Returns
    -------
    inp_to_counts : DataArray or Dataset
        Array with the same shape as the input array, where each element
        is the number of non-NaN values at the corresponding time step.

    """
    inp_to_counts = xr.where(np.isnan(inp), 0, 1)

    return inp_to_counts


def feeps_avg_4sc(
    b_list: Sequence[DataArray], flag: str = "omni", combined_energies: list = None
) -> DataArray:
    r"""Average data from 4 spacecrafts while ignoring NaN values.
    Computes the input quantity at the center of mass of the MMS
    tetrahedron. When averaging, NaN values are ignored by counting the number of
    non-NaN values at each time step.

    Parameters
    ----------
    b_list : Sequence of DataArray or Dataset
        List of the time series of the quantity for each spacecraft.

    flag : str, optional
        type of data to be averaged (omnidirectional or
        pitch-angle distribution). The default is "omni".

    combined_energies : list, optional
        Used only if flag is "omni".
        List of energy channel indices from each spacecraft to be combined into a common
        energy channel. The default is None, which means that the initial channels from
        the omnidirectional data are used. This will likely result to averaging fluxes
        for much different energies.

    Returns
    -------
    b_avg : DataArray or Dataset
        Time series of the input quantity at the center of mass of the
        MMS tetrahedron.

    Raises
    ------
    TypeError
        If b_list is not a list of DataArray or Dataset

    """
    # Check input type
    if not isinstance(b_list, list):
        raise TypeError("b_list must be a list")

    assert flag in ["omni", "pad"], "flag must be 'omni' or 'pad'"

    b_list_r = []

    for b in b_list:
        if isinstance(b, (xr.DataArray, xr.Dataset)):
            b_list_r.append(resample(b, b_list[0], f_s=calc_fs(b_list[0])))
        else:
            raise TypeError("elements of b_list must be DataArray or Dataset")

    b_list_count_nans = deepcopy(b_list_r)
    b_list_r = [xr.where(np.isnan(b), 0, b) for b in b_list_r]
    b_avg_data = np.zeros(b_list_r[0].shape)
    b_nan_denom = np.zeros(b_list_r[0].shape)

    if flag == "omni" and combined_energies is not None:
        if not isinstance(combined_energies, dict):
            raise TypeError("combined_energies must be a dict")

        energy = np.zeros(len(b_list_r[0].energy.data))

        for ien_bin in range(len(b_list_r[0].energy.data)):
            non_none_sc_ch = 0
            for i, b, b_count_nan in zip(
                range(len(b_list_r)), b_list_r, b_list_count_nans
            ):
                mms_id = b_list[i].attrs["mmsId"] - 1
                if combined_energies[ien_bin][mms_id] is None:
                    continue
                non_none_sc_ch += 1
                b_avg_data[:, ien_bin] += b.data[
                    :, combined_energies[ien_bin][mms_id] - 1
                ]
                b_nan_denom[:, ien_bin] += _nan_count(
                    b_count_nan.data[:, combined_energies[ien_bin][mms_id] - 1]
                ).data
                energy[ien_bin] += b.energy.data[combined_energies[ien_bin][mms_id] - 1]
            energy[ien_bin] /= non_none_sc_ch
    else:

        for b, b_count_nan in zip(b_list_r, b_list_count_nans):

            b_avg_data += b.data
            b_nan_denom += _nan_count(b_count_nan).data

    coords = {}
    if flag == "omni":
        if combined_energies is None:
            energy = b_list[0].energy.data
        coords = {"time": b_list_r[0].time, "energy": energy}
    elif flag == "pad":
        coords = {"time": b_list_r[0].time, "theta": b_list_r[0].theta}

    b_avg = xr.DataArray(
        b_avg_data / b_nan_denom,
        coords=coords,
        dims=b_list_r[0].dims,
        attrs=b_list[0].attrs,
    )

    if "probe" in b_avg.attrs.keys():
        b_avg.attrs["probe"] = "4sc_avg"
    if "mms" in b_avg.attrs.keys():
        b_avg.attrs["mms"] = "4sc_avg"
    if "MMS" in b_avg.attrs.keys():
        b_avg.attrs["MMS"] = "4sc_avg"
    if "mmsId" in b_avg.attrs.keys():
        b_avg.attrs["mmsId"] = "4sc_avg"

    return b_avg



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# Built-in imports
import warnings
from datetime import datetime, timedelta

# 3rd party imports
import numpy as np

# Local imports
from .list_files_sdc import _login_lasp

__author__ = "Louis Richard"
__email__ = "louisr@irfu.se"
__copyright__ = "Copyright 2020-2023"
__license__ = "MIT"
__version__ = "2.4.11"
__status__ = "Prototype"


def _construct_url_json_list(tint, mms_id, product, lasp_url):
    r"""Construct the url that return a json-formatted string of science
    filenames that are available for download according to:
    https://lasp.colorado.edu/mms/sdc/team/about/how-to/
    """

    tint = np.array(tint).astype("<M8[ns]").astype(str)
    tint = [datetime.strptime(t_[:-3], "%Y-%m-%dT%H:%M:%S.%f") for t_ in tint]
    start_date = (tint[0] - timedelta(days=1)).strftime("%Y-%m-%d")
    end_date = (tint[1] + timedelta(days=1)).strftime("%Y-%m-%d")

    url = f"{lasp_url}/file_info/ancillary"
    url = f"{url}?start_date={start_date}&end_date={end_date}&sc_id=mms{mms_id}"

    url = f"{url}&product={product}"

    return url


def _make_urls_ancillaries(lasp_url, files):
    for file in files:
        file["url"] = f"{lasp_url}download/ancillary?file={file['file_name']}"

    return files


def list_files_ancillary_sdc(tint, mms_id, product):
    r"""Find available ancillary files from LASP SDC for the target product type.

    Parameters
    ----------
    tint : list of str
        Time interval
    mms_id : str or int
        Spacecraft index
    product : {"predatt", "predeph", "defatt", "defeph"}
        Ancillary type.

    Returns
    -------
    file_names : list
        Ancillary files in interval.

    """

    sdc_session, headers, lasp_url = _login_lasp()

    url_json_ancillaries = _construct_url_json_list(tint, mms_id, product, lasp_url)

    with warnings.catch_warnings():
        warnings.simplefilter("ignore", category=ResourceWarning)
        response = sdc_session.get(url_json_ancillaries, verify=True, headers=headers)
        response.raise_for_status()  # Raise an HTTPError for bad responses
        http_json = response.json()

    file_names = http_json["files"]
    sdc_session.close()

    file_names = _make_urls_ancillaries(lasp_url, file_names)

    return file_names



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# Third party imports
import numpy as np

# Built-in imports
import tqdm
import xarray as xr
from scipy.constants import electron_mass, electron_volt, proton_mass, speed_of_light

# Local imports
from pyrfu.pyrf.datetime642iso8601 import datetime642iso8601
from pyrfu.pyrf.int_sph_dist import int_sph_dist
from pyrfu.pyrf.resample import resample
from pyrfu.pyrf.time_clip import time_clip
from pyrfu.pyrf.ts_scalar import ts_scalar

__author__ = "Louis Richard"
__email__ = "louisr@irfu.se"
__copyright__ = "Copyright 2020-2023"
__license__ = "MIT"
__version__ = "2.4.2"
__status__ = "Prototype"


def reduce(vdf, xyz, dim: str = "1d", base: str = "pol", **kwargs):
    r"""Reduces (integrates) 3D distribution to 1D (line) or 2D (plane).
    Draft do not use!!

    Parameters
    ----------
    vdf : xarray.Dataset
        3D skymap velocity distribution function.
    xyz : xarray.DataArray or numpy.ndarray
        Transformation matrix from instrument frame to desired frame.
    base : str, Optional
        Base for the 2D projection either cartesian 'cart' (default) or
        polar 'pol'.
    **kwargs
        Keyword arguments

    Returns
    -------
    out : xarray.DataArray
        Time series of reduced velocity distribution function.

    """

    # Make sure the projection dimension amd base are correct
    assert base.lower() in ["cart", "pol"], "Invalid projection base!!"
    assert dim.lower() in ["1d", "2d", "3d"], "Invalid projection dimension!!"
    ndim = int(dim[0])

    # Clip the distribution. If no time interval provided use the entire
    # time series.
    # Note: irfu-matlab does the rotation matrix first and then time clip but here
    # we do time clip first to save computation time if the time interval is
    # smaller than the VDF time series.
    vdf_time = vdf.time
    tint = kwargs.get("tint", list(datetime642iso8601(vdf_time.data[[0, -1]])))
    vdf_time = time_clip(vdf_time, tint)
    vdf_energy = time_clip(vdf.energy, tint).copy()
    vdf_phi = time_clip(vdf.phi, tint).copy()
    vdf_theta = vdf.theta.copy()
    vdf_data = time_clip(vdf.data, tint).copy()

    # Lower energy bound of instrument bins. If not provided set to 0 eV.
    if "delta_energy_minus" in vdf.attrs:
        delta_energy_minu = xr.DataArray(
            vdf.attrs["delta_energy_minus"],
            coords=[vdf.time.data, vdf.idx0.data],
            dims=["time", "idx0"],
        )
    else:
        delta_energy_minu = xr.DataArray(
            np.zeros((len(vdf.time.data), len(vdf.idx0.data))),
            coords=[vdf.time.data, vdf.idx0.data],
            dims=["time", "idx0"],
        )

    delta_energy_minu = time_clip(delta_energy_minu, tint)

    # make input distribution to SI units, s^3/m^6
    if vdf.data.attrs["UNITS"].lower() == "s^3/cm^6":
        vdf_data *= 1e12
    elif vdf.data.attrs["UNITS"].lower() == "s^3/m^6":
        vdf_data *= 1e0
    elif vdf.data.attrs["UNITS"].lower() == "s^3/km^6":
        vdf_data *= 1e-18
    else:
        raise ValueError("Invalid units!!")

    # Get VDF dimension (time, energy, phi, theta)
    # n_t, _, n_ph, _ = vdf_data.shape
    n_t, n_en, _, _ = vdf_data.shape

    # Construct or resample the time series of transformation matrix
    if isinstance(xyz, xr.DataArray):
        # If time series, clip to time interval and resample to VDF's time line
        xyz = time_clip(xyz, tint)
        xyz = resample(xyz, vdf_time).data
    elif isinstance(xyz, np.ndarray) and xyz.ndim == 2:
        assert xyz.shape == (3, 3), "xyz must be a transformation matrix!!"
        # If matrix, tile to VDF timeline
        xyz = np.tile(xyz, (n_t, 1, 1))
    else:
        raise TypeError("Invalid type for xyz")

    # Check that the transformation matrices are orthonormal direct
    x_phat_ts, y_phat_ts, z_phat_ts = [xyz[:, :, i] for i in range(3)]
    x_phat_ts /= np.linalg.norm(x_phat_ts, axis=1, keepdims=True)
    y_phat_ts /= np.linalg.norm(y_phat_ts, axis=1, keepdims=True)
    z_phat_ts = np.cross(x_phat_ts, y_phat_ts)
    z_phat_ts /= np.linalg.norm(z_phat_ts, axis=1, keepdims=True)
    y_phat_ts = np.cross(z_phat_ts, x_phat_ts)

    # Construct the time series of transformation matrix
    xyz_ts = np.transpose(
        np.stack([x_phat_ts, y_phat_ts, z_phat_ts]),
        [1, 2, 0],
    )

    # This is now done in int_sph_dist
    # Set azimuthal angle projection grid.
    # d_phi = kwargs.get("d_phi", 2 * np.pi / n_ph)
    # phi_grid = np.linspace(0, 2 * np.pi, n_ph) + d_phi / 2  # centers

    # Read the keyword arguments
    # Velocity projection grid and edges
    velocity_grid = kwargs.get("vg", None)  # TODO : check that for no input!!
    velocity_grid_edges = kwargs.get("vg_edges", None)

    # azimuthal angle of projection plane
    n_phi_grid = len(vdf_phi)
    d_phi_g = 2 * np.pi / n_phi_grid
    phi_grid = np.linspace(0, 2 * np.pi - d_phi_g, n_phi_grid) + d_phi_g / 2
    phi_grid = kwargs.get("phig", phi_grid)

    # Number of Monte Carlo iterations and weights
    n_mc = kwargs.get("n_mc", 100)  # OK
    weight = kwargs.get("weight", None)  # OK

    # Velocity and azimuthal angle integration intervals
    v_int = kwargs.get("v_int", [-np.inf, np.inf])  # OK
    a_int = kwargs.get("a_int", [-180.0, 180.0])  # OK

    # Spacecraft potential to account for photoelectrons. If not provided set
    # to 0 V.
    sc_pot = kwargs.get("sc_pot", ts_scalar(vdf_time.data, np.zeros(n_t)))
    sc_pot = resample(sc_pot, vdf_time).data

    # Threshold energy (e.g., to remove background noise). If not provided set
    # to 0 eV.
    lower_e_lim = kwargs.get("lower_e_lim", 0.0)

    if isinstance(lower_e_lim, xr.DataArray):
        # If time series, clip to time interval and resample to VDF's time line
        lower_e_lim = resample(lower_e_lim, vdf_time).data
    elif isinstance(lower_e_lim, float):
        # If float, tile to VDF's timeline
        lower_e_lim = np.tile(lower_e_lim, n_t)
    else:
        raise TypeError("Invalid lower_e_lim!!")

    # Set particle specie mass from VDF's attribute "species"
    if vdf.species.lower() == "electrons":
        m_p = electron_mass
    elif vdf.species.lower() == "ions":
        m_p = proton_mass
    else:
        raise ValueError("Invalid species!!")

    # Convert maximum energy from instrument to velocity (relativistically
    # correct)
    e_max = vdf_energy.data[0, -1] + vdf.attrs["delta_energy_plus"][0, -1]
    gamma_max = 1 + electron_volt * e_max / (m_p * speed_of_light**2)
    v_max = speed_of_light * np.sqrt(1 - 1 / gamma_max**2)  # m/s

    # initiate projected f
    if velocity_grid_edges is not None:
        n_vg = len(velocity_grid_edges) - 1
        velocity_grid = velocity_grid_edges[:-1] + 0.5 * np.diff(velocity_grid_edges)
    elif velocity_grid is not None:
        n_vg = len(velocity_grid)
    else:
        n_vg = 100
        if base == "cart":
            n_vg = 100
        elif base == "pol":
            if ndim == 1:
                n_vg = 2 * n_en
            else:
                n_vg = n_en

    # Initialize output arrays
    f_g = np.zeros([n_t, *[n_vg] * ndim])
    all_v = {f"v{chr(120 + i)}": np.zeros((n_t, n_vg)) for i in range(ndim)}

    for i_t in tqdm.tqdm(range(n_t), ncols=60):  # display progress
        # 3d data matrix for time index
        f_3d = np.squeeze(vdf_data.data[i_t, ...])  # s^3/m^6
        f_3d = f_3d.astype(np.float64).copy()  # Convert to float64

        # Energies
        energy = vdf_energy.data[i_t, :]
        energy = energy.astype(np.float64)  # Convert to float64

        # Assign zeros to phase-space density values corresponding to energies
        # below the threshold energy or below the spacecraft potential if
        # provided.
        thresh_e = np.max([lower_e_lim[i_t], sc_pot[i_t]])
        e_min_idx = np.where(energy - delta_energy_minu[i_t, :] > thresh_e)[0][0]
        f_3d[:e_min_idx] = 0.0

        # Correct energy shift due to spacecraft potential
        energy -= sc_pot[i_t]
        energy[energy < 0] = 0.0

        # Convert energy to velocity (relativistically correct)
        gamma = 1 + electron_volt * energy / (m_p * speed_of_light**2)
        velocity = speed_of_light * np.sqrt(1 - 1 / gamma**2)  # m/s

        # azimuthal angle
        if vdf_phi.ndim == 2:
            phi = vdf_phi.data[i_t, :].astype(np.float64)  # in degrees
        else:
            # Fast mode
            phi = vdf_phi.data.astype(np.float64)  # in degrees

        # Remove 180 deg to convert from instrument to physical frame
        phi = np.deg2rad(phi - 180.0)  # in radians

        # elevation angle and remove 90 deg to convert from instrument to
        # physical frame.
        theta = vdf_theta.data.astype(np.float64)  # in degrees
        theta = np.deg2rad(theta - 90.0)  # in radians

        # TODO: add delta_phi and delta_theta if needed

        # Set velocity projection grid.
        if velocity_grid is None:
            if base == "cart":
                velocity_grid = np.linspace(-v_max, v_max, 100, endpoint=True)
            elif base == "pol":
                if ndim == 1:
                    velocity_grid = np.hstack((-np.flip(velocity), velocity))
                else:
                    velocity_grid = velocity

        else:
            pass

        options = {
            "xyz": xyz_ts[i_t, ...],
            "n_mc": n_mc,
            "weight": weight,
            "v_int": v_int,
            "a_int": a_int,
            "projection_dim": dim,
            "projection_base": base,
            "velocity_grid_edges": velocity_grid_edges,
        }

        tmpst = int_sph_dist(
            f_3d, velocity, phi, theta, velocity_grid, phi_grid, **options
        )

        f_g[i_t, ...] = tmpst["f"]

        for i in range(ndim):
            all_v[f"v{chr(120 + i)}"][i_t, :] = tmpst[f"v{chr(120 + i)}"] / 1e3  # km/s

    # Build output as a time series with dimensions:
    #   - (time x vx) for 1D reduced distribution
    #   - (time x vx x vy) for 2D reduced distribution
    coords = [
        vdf_time.data,
        *[all_v[f"v{chr(120 + i)}"][0, :] for i in range(ndim)],
    ]
    dims = ["time", *[f"v{chr(120 + i)}" for i in range(ndim)]]
    out = xr.DataArray(f_g, coords=coords, dims=dims)

    return out



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# 3rd party imports
import numpy as np
from xarray.core.dataarray import DataArray

# Local imports
from pyrfu.pyrf.resample import resample
from pyrfu.pyrf.ts_vec_xyz import ts_vec_xyz

__author__ = "Louis Richard"
__email__ = "louisr@irfu.se"
__copyright__ = "Copyright 2020-2024"
__license__ = "MIT"
__version__ = "2.4.13"
__status__ = "Prototype"


def cross(inp1: DataArray, inp2: DataArray) -> DataArray:
    r"""Compute cross product of two fields.

    Parameters
    ----------
    inp1 : DataArray
        Time series of the first field X.
    inp2 : DataArray
        Time series of the second field Y.

    Returns
    -------
    DataArray
        Time series of the cross product Z = XxY.

    Raises
    ------
    TypeError
        If inp1 or inp2 are not xarray.DataArray.
    ValueError
        If inp1 or inp2 are not vectors timeseries.

    Examples
    --------
    >>> from pyrfu import mms, pyrf

    Define time interval

    >>> tint = ["2019-09-14T07:54:00.000", "2019-09-14T08:11:00.000"]

    Index of the MMS spacecraft

    >>> mms_id = 1

    Load magnetic field and electric field

    >>> b_xyz = mms.get_data("b_gse_fgm_srvy_l2", tint, mms_id)
    >>> e_xyz = mms.get_data("e_gse_edp_fast_l2", tint, mms_id)

    Compute magnitude of the magnetic field

    >>> b_mag = pyrf.norm(b_xyz)

    Compute ExB drift velocity

    >>> v_xyz_exb = pyrf.cross(e_xyz, b_xyz) / b_mag ** 2

    """
    # Check type
    if not isinstance(inp1, DataArray):
        raise TypeError("inp1 must be a xarray.DataArray")

    if not isinstance(inp2, DataArray):
        raise TypeError("inp2 must be a xarray.DataArray")

    # Check inputs are vectors
    if inp1.ndim != 2 or inp1.shape[1] != 3:
        raise ValueError("inp1 must be a vector")

    if inp2.ndim != 2 or inp2.shape[1] != 3:
        raise ValueError("inp2 must be a vector")

    if len(inp1) != len(inp2):
        inp2 = resample(inp2, inp1)

    out_data = np.cross(inp1.data, inp2.data, axis=1)

    out = ts_vec_xyz(inp1.time.data, out_data)

    return out



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# 3rd party imports
import numpy as np
from xarray.core.dataarray import DataArray

__author__ = "Louis Richard"
__email__ = "louisr@irfu.se"
__copyright__ = "Copyright 2020-2023"
__license__ = "MIT"
__version__ = "2.4.2"
__status__ = "Prototype"


def normalize(inp: DataArray) -> DataArray:
    r"""Normalizes the input field.

    Parameters
    ----------
    inp : DataArray
        Time series of the input field.

    Returns
    -------
    DataArray
        Time series of the normalized input field.

    Raises
    ------
    TypeError
        If input is not a DataArray.
    ValueError
        If input is not a 2D DataArray.

    Examples
    --------
    >>> from pyrfu import mms, pyrf

    Time interval

    >>> tint = ["2019-09-14T07:54:00.000", "2019-09-14T08:11:00.000"]

    Spacecraft index

    >>> mms_id = 1

    Load magnetic field

    >>> b_xyz = mms.get_data("B_gse_fgm_srvy_l2", tint, mms_id)

    Compute the normalized magnetic field

    >>> b = pyrf.normalize(b_xyz)

    """

    if not isinstance(inp, DataArray):
        raise TypeError("Input must be a DataArray")

    if inp.data.ndim == 2:
        out_data = inp.data / np.linalg.norm(inp.data, axis=1, keepdims=True)
        out = inp.copy(data=out_data)
    else:
        raise ValueError("Input must be a 2D DataArray")

    return out



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# 3rd party imports
import numpy as np
import pycdfpp
import xarray as xr

__author__ = "Louis Richard"
__email__ = "louisr@irfu.se"
__copyright__ = "Copyright 2020-2023"
__license__ = "MIT"
__version__ = "2.4.13"
__status__ = "Prototype"


def read_cdf(path: str) -> dict:
    r"""Reads a .cdf file and returns a dictionary with the fields contained in
    the file.

    Parameters
    ----------
    path : str
        Path to the .cdf file.

    Returns
    -------
    dict
        Hash table with fields contained in the .cdf file.

    """

    # Initialize output dictionary
    out_dict = {}

    # Load file
    file = pycdfpp.load(path)

    # Get keys (a.k.a zvariables) from file
    keys = list(map(lambda x: x[0], file.items()))

    for key in keys:
        # Get data and coordinates
        data = np.squeeze(file[key].values)
        coords = [np.arange(dim_size) for dim_size in data.shape]

        # Construct xarray DataArray
        out_dict[key.lower()] = xr.DataArray(data, coords=coords)

    return out_dict



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# Built-in imports
from typing import Any, Iterable, Mapping, Optional, Union

# 3rd party imports
import numpy as np
import xarray as xr
from numpy.typing import NDArray
from xarray.core.dataset import Dataset

__author__ = "Louis Richard"
__email__ = "louisr@irfu.se"
__copyright__ = "Copyright 2020-2024"
__license__ = "MIT"
__version__ = "2.4.13"
__status__ = "Prototype"

NDArrayFloats = NDArray[Union[np.float32, np.float64]]


def ts_skymap(
    time: NDArray[np.datetime64],
    data: NDArrayFloats,
    energy: NDArrayFloats,
    phi: NDArrayFloats,
    theta: NDArrayFloats,
    energy0: Optional[NDArrayFloats] = None,
    energy1: Optional[NDArrayFloats] = None,
    esteptable: Optional[NDArray[np.uint8]] = None,
    attrs: Optional[Mapping[str, object]] = None,
    glob_attrs: Optional[Mapping[str, object]] = None,
    coords_attrs: Optional[Mapping[str, Mapping[str, Iterable[Any]]]] = None,
) -> Dataset:
    r"""Creates a skymap of the distribution function.

    Parameters
    ----------
    time : np.ndarray
        List of times.
    data : np.ndarray
        Values of the distribution function.
    energy : np.ndarray
        Energy levels.
    phi : np.ndarray
        Azimuthal angles.
    theta : np.ndarray
        Elevation angles.
    energy0: np.ndarray, Optional
        Energy table 0 (odd time indices).
    energy1: np.ndarray, Optional
        Energy table 1 (even time indices).
    esteptable: np.ndarray, Optional
        Time series of the stepping table between energies (burst).
    attrs : dict, Optional
        Metadata for the VDF.
    glob_attrs : dict, Optional
        Global attributes of the dataset.
    coords_attrs : dict, Optional
        Coordinates attributes of the dataset.

    Returns
    -------
    out : xarray.Dataset
        Skymap of the distribution function.

    Raises
    ------
    TypeError
        If time, data, energy, phi, or theta are not numpy.ndarray.
    TypeError
        If energy0, energy1, or esteptable are not numpy.ndarray.
    TypeError
        If attrs, glob_attrs, or coords_attrs are not dict.

    """
    # Check input type
    if not isinstance(time, np.ndarray):
        raise TypeError("time must be a numpy.ndarray")

    if not isinstance(data, np.ndarray):
        raise TypeError("data must be a numpy.ndarray")

    if not isinstance(energy, np.ndarray):
        raise TypeError("energy must be a numpy.ndarray")

    if not isinstance(phi, np.ndarray):
        raise TypeError("phi must be a numpy.ndarray")

    if not isinstance(theta, np.ndarray):
        raise TypeError("theta must be a numpy.ndarray")

    # Check if even (odd) time step energy channels energy1 (energy0), and
    # energy step table are provided.
    if energy0 is None:
        energy0 = energy[0, :]
    else:
        if not isinstance(energy0, np.ndarray):
            raise TypeError("energy0 must be a numpy.ndarray")

    if energy1 is None:
        energy1 = energy[1, :]
    else:
        if not isinstance(energy1, np.ndarray):
            raise TypeError("energy0 must be a numpy.ndarray")

    if esteptable is None:
        esteptable = np.zeros(len(time), dtype=np.uint8)
    else:
        if not isinstance(esteptable, np.ndarray):
            raise TypeError("esteptable must be a numpy.ndarray")

    # Check attributes are dictionaries
    if attrs is None:
        attrs = {}
    else:
        if not isinstance(attrs, dict):
            raise TypeError("attrs must be a dictionary")

    # Check coordinates attributes are dictionaries
    if coords_attrs is None:
        coords_attrs = {}
    else:
        if not isinstance(coords_attrs, dict):
            raise TypeError("coords_attrs must be a dictionary")

    # Check global attributes are dictionaries
    if glob_attrs is None:
        glob_attrs = {}
    else:
        if not isinstance(glob_attrs, dict):
            raise TypeError("glob_attrs must be a dictionary")

    out_dict = {
        "data": (["time", "idx0", "idx1", "idx2"], data),
        "phi": (["time", "idx1"], phi),
        "theta": (["idx2"], theta),
        "energy": (["time", "idx0"], energy),
        "time": time,
        "idx0": np.arange(energy.shape[1]),
        "idx1": np.arange(phi.shape[1]),
        "idx2": np.arange(len(theta)),
    }

    # Construct global attributes and sort them
    # remove energy0, energy1, and esteptable from global attrs to overwrite
    overwrite_keys = ["energy0", "energy1", "esteptable"]
    glob_attrs = {k: glob_attrs[k] for k in glob_attrs if k not in overwrite_keys}
    glob_attrs = {
        "energy0": energy0,
        "energy1": energy1,
        "esteptable": esteptable,
        **glob_attrs,
    }

    glob_attrs = {k: glob_attrs[k] for k in sorted(glob_attrs)}

    # Create Dataset
    out = xr.Dataset(out_dict, attrs=glob_attrs)

    # Sort and fill coordinates attributes
    for k in coords_attrs:
        out[k].attrs = {m: coords_attrs[k][m] for m in sorted(coords_attrs[k])}

    # Sort and fill data attributes
    out.data.attrs = {k: attrs[k] for k in sorted(attrs)}

    return out



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# 3rd party imports
import numpy as np
import xarray as xr

# Local imports
from .ts_vec_xyz import ts_vec_xyz

__author__ = "Louis Richard"
__email__ = "louisr@irfu.se"
__copyright__ = "Copyright 2020-2023"
__license__ = "MIT"
__version__ = "2.4.2"
__status__ = "Prototype"


def cart2sph_ts(inp, direction_flag: int = 1):
    r"""Computes magnitude, theta and phi angle from column vector xyz
    (first column is x ....) theta is 0 at equator.
    direction_flag = -1  -> to make transformation in opposite direction

    Parameters
    ----------
    inp : xarray.DataArray
        Time series to convert.

    direction_flag : {1, -1}, Optional
        Set to 1 (default) to transform from cartesian to spherical
        coordinates. Set to -1 to transform from spherical to cartesian
        coordinates.

    Returns
    -------
    out : xarray.DataArray
        Input field in spherical/cartesian coordinate system.

    """

    # Check input type
    assert isinstance(inp, xr.DataArray), "inp must be a xarray.DataArray"

    # Check that inp is a vector time series
    assert inp.data.ndim == 2 and inp.shape[1] == 3, "inp must be a vector time series"

    # Check direction +/-1
    assert direction_flag in [-1, 1], "direction_flag must be +/-1"

    if direction_flag == -1:
        r_data = inp.data[:, 0]

        sin_the = np.sin(np.deg2rad(inp.data[:, 1]))
        cos_the = np.cos(np.deg2rad(inp.data[:, 1]))
        sin_phi = np.sin(np.deg2rad(inp.data[:, 2]))
        cos_phi = np.cos(np.deg2rad(inp.data[:, 2]))

        z_data = r_data * sin_the
        x_data = r_data * cos_the * cos_phi
        y_data = r_data * cos_the * sin_phi

        out_data = np.transpose(np.vstack([x_data, y_data, z_data]))

    else:
        xy2 = inp.data[:, 0] ** 2 + inp.data[:, 1] ** 2

        r_data = np.sqrt(xy2 + inp.data[:, 2] ** 2)
        theta = np.rad2deg(np.arctan2(inp.data[:, 2], np.sqrt(xy2)))
        phi = np.rad2deg(np.arctan2(inp.data[:, 1], inp.data[:, 0]))

        out_data = np.transpose(np.vstack([r_data, theta, phi]))

    out = ts_vec_xyz(inp.time.data, out_data, inp.attrs)

    return out



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# Built-in imports
from typing import Optional

# 3rd party imports
import numpy as np
import xarray as xr
from xarray.core.dataarray import DataArray

# Local imports
from pyrfu.pyrf.calc_dt import calc_dt

__author__ = "Louis Richard"
__email__ = "louisr@irfu.se"
__copyright__ = "Copyright 2020-2024"
__license__ = "MIT"
__version__ = "2.4.13"
__status__ = "Prototype"


def autocorr(
    inp: DataArray, maxlags: Optional[int] = None, normed: Optional[bool] = True
) -> DataArray:
    r"""Compute the autocorrelation function.

    Parameters
    ----------
    inp : DataArray
        Input time series (scalar of vector).
    maxlags : int, Optional
        Maximum lag in number of points. Default is None (i.e., len(inp) - 1).
    normed : bool, Optional
        Flag to normalize the correlation.

    Returns
    -------
    DataArray
        Autocorrelation function

    Raises
    ------
    TypeError
        If inp is not a xarray.DataArray
    ValueError
        If inp is not a scalar or a vector
    ValueError
        If maxlags is not None or strictly positive < len(inp)

    """
    # Check input type
    if not isinstance(inp, DataArray):
        raise TypeError("inp must be a DataArray")

    # Check input dimension (scalar or vector)
    if inp.ndim > 2:
        raise ValueError("inp must be a scalar or a vector")

    if inp.ndim == 1:
        x = inp.data[:, None]
    else:
        x = inp.data

    n_t = len(inp)

    if maxlags is None:
        maxlags = n_t - 1

    if maxlags >= n_t or maxlags < 1:
        raise ValueError(f"maxlags must be None or strictly positive < {n_t:d}")

    lags = np.linspace(-float(maxlags), float(maxlags), 2 * maxlags + 1, dtype=int)
    lags = lags * calc_dt(inp)

    out_data = np.zeros((maxlags + 1, x.shape[1]))

    for i in range(x.shape[1]):
        correls = np.correlate(x[:, i], x[:, i], mode="full")

        if normed:
            correls /= np.sqrt(np.dot(x[:, i], x[:, i]) ** 2)

        correls = correls[n_t - 1 - maxlags : n_t + maxlags]
        out_data[:, i] = correls[lags >= 0]

    if inp.ndim == 1:
        coords = [lags[lags >= 0]]
        dims = ["lag"]
    else:
        coords = [lags[lags >= 0], inp[inp.dims[1]].data]
        dims = ["lag", str(inp.dims[1])]

    return xr.DataArray(np.squeeze(out_data), coords=coords, dims=dims)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# 3rd party imports
import numpy as np
import xarray as xr

__author__ = "Louis Richard"
__email__ = "louisr@irfu.se"
__copyright__ = "Copyright 2020-2023"
__license__ = "MIT"
__version__ = "2.4.2"
__status__ = "Prototype"


def mva(inp, flag: str = "mvar"):
    r"""Compute the minimum variance frame.

    Parameters
    ----------
    inp : xarray.DataArray
        Time series of the quantity to find minimum variance frame.
    flag : {"mvar", "<bn>=0", "td"}, Optional
        Constrain. Default is "mvar".

    Returns
    -------
    out : xarray.DataArray
        Time series of the input quantity in LMN coordinates.
    l : numpy.ndarray
        Eigenvalues l[0] > l[1] > l[2].
    lmn : numpy.ndarray
        Eigenvectors LMN coordinates.

    See also
    --------
    pyrfu.pyrf.new_xyz

    Examples
    --------
    >>> from pyrfu import mms, pyrf

    Time interval

    >>> tint = ["2019-09-14T07:54:00.000", "2019-09-14T08:11:00.000"]

    Spacecraft index

    >>> mms_id = 1

    Load magnetic field

    >>> b_xyz = mms.get_data("B_gse_fgm_srvy_l2", tint, mms_id)

    Compute MVA frame

    >>> b_lmn, lamb, frame = pyrf.mva(b_xyz)

    """

    assert flag.lower() in ["mvar", "<bn>=0", "td"], "invalid method!!"

    inp_data = inp.data
    n_t = inp_data.shape[0]

    idx_1, idx_2 = [[0, 1, 2, 0, 0, 1], [0, 1, 2, 1, 2, 2]]

    if flag in ["mvar", "<bn>=0"]:
        m_mu_nu_m = np.mean(inp_data[:, idx_1] * inp_data[:, idx_2], 0)
        m_mu_nu_m -= np.mean(inp_data, 0)[idx_1] * np.mean(inp_data, 0)[idx_2]

    else:
        m_mu_nu_m = np.mean(inp_data[:, idx_1] * inp_data[:, idx_2], 0)

    m_mu_nu = np.array(
        [m_mu_nu_m[[0, 3, 4]], m_mu_nu_m[[3, 1, 5]], m_mu_nu_m[[4, 5, 2]]],
    )

    # Compute eigenvalues and eigenvectors
    [lamb, lmn] = np.linalg.eig(m_mu_nu)

    # Sort eigenvalues
    lamb, lmn = [lamb[lamb.argsort()[::-1]], lmn[:, lamb.argsort()[::-1]]]

    # ensure that the frame is right handed
    lmn[:, 2] = np.cross(lmn[:, 0], lmn[:, 1])

    if flag.lower() == "<bn>=0":
        inp_mvar_mean = np.mean(
            np.sum(
                np.tile(inp_data, (3, 1, 1))
                * np.transpose(np.tile(lmn, (n_t, 1, 1)), (2, 0, 1)),
                1,
            ),
            1,
        )

        coeffs = np.zeros(3)
        coeffs[0] = np.sum(inp_mvar_mean**2)

        coeffs[1] = -(lamb[1] + lamb[2]) * inp_mvar_mean[0] ** 2
        coeffs[1] -= (lamb[0] + lamb[2]) * inp_mvar_mean[1] ** 2
        coeffs[1] -= (lamb[0] + lamb[1]) * inp_mvar_mean[2] ** 2

        coeffs[2] = lamb[1] * lamb[2] * inp_mvar_mean[0] ** 2
        coeffs[2] += lamb[0] * lamb[2] * inp_mvar_mean[1] ** 2
        coeffs[2] += lamb[0] * lamb[1] * inp_mvar_mean[2] ** 2

        roots = np.roots(coeffs)

        l_min = np.min(roots)

        n_vec = inp_mvar_mean / (lamb - l_min)
        n_vec /= np.linalg.norm(n_vec, keepdims=True)
        n_vec = np.matmul(lmn, n_vec)

        b_vec = np.sum(inp_data * np.tile(n_vec, (n_t, 1)), axis=1)

        inp_data_2 = inp_data.copy()
        inp_data_2 -= np.tile(b_vec, (3, 1)).T * np.tile(n_vec, (n_t, 1))

        inp_data_2_m = np.mean(inp_data_2, 0)

        m_mu_nu_m = np.mean(inp_data_2[:, idx_1] * inp_data_2[:, idx_2], 0)
        m_mu_nu_m -= inp_data_2_m[idx_1] * inp_data_2_m[idx_2]

        m_mu_nu = np.array(
            [m_mu_nu_m[[0, 3, 4]], m_mu_nu_m[[3, 1, 5]], m_mu_nu_m[[4, 5, 2]]],
        )

        lamb, lmn = np.linalg.eig(m_mu_nu)

        lamb, lmn = [lamb[lamb.argsort()[::-1]], lmn[:, lamb.argsort()[::-1]]]

        # Force the maximum variance direction to be positive
        # lmn[:, 0] *= np.sign(lmn[np.argmax(lmn[:, 0]), 0])
        # lamb[2], lmn[:, 2] = [l_min, np.cross(lmn[:, 0], lmn[:, 1])]

    elif flag.lower() == "td":
        l_min = lamb[2]
        b_vec = np.sum(inp_data * np.tile(lmn[:, 2], (n_t, 1)), axis=1)

        inp_data_2 = inp_data.copy()
        inp_data_2 -= np.tile(b_vec, (3, 1)).T * np.tile(lmn[:, 2], (n_t, 1))

        inp_data_2_m = np.mean(inp_data_2, 0)

        m_mu_nu_m = np.mean(inp_data_2[:, idx_1] * inp_data_2[:, idx_2], 0)
        m_mu_nu_m -= inp_data_2_m[idx_1] * inp_data_2_m[idx_2]

        m_mu_nu = np.array(
            [m_mu_nu_m[[0, 3, 4]], m_mu_nu_m[[3, 1, 5]], m_mu_nu_m[[4, 5, 2]]],
        )

        lamb, lmn = np.linalg.eig(m_mu_nu)

        lamb, lmn = [lamb[lamb.argsort()[::-1]], lmn[:, lamb.argsort()[::-1]]]

        lamb[2], lmn[:, 2] = [l_min, np.cross(lmn[:, 0], lmn[:, 1])]
    else:
        pass

    out_data = (lmn.T @ inp_data.T).T

    out = xr.DataArray(out_data, coords=inp.coords, dims=inp.dims)

    return out, lamb, lmn



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# 3rd party imports
import numpy as np
import xarray as xr
from xarray.core.dataarray import DataArray

# Local imports
from pyrfu.pyrf.ts_scalar import ts_scalar

__author__ = "Louis Richard"
__email__ = "louisr@irfu.se"
__copyright__ = "Copyright 2020-2024"
__license__ = "MIT"
__version__ = "2.4.13"
__status__ = "Prototype"


def calc_ag(p_xyz: DataArray) -> DataArray:
    r"""Compute Che's agyrotropy coefficient.

     Che's agyrotropy is [16]_

    .. math::

        AG^{1/3} = \frac{|\operatorname[det]{\mathbf{P}}
        - \operatorname[det]{\mathbf{P}}|}
        {\operatorname[det]{\mathbf{P}}
        + \operatorname[det]{\mathbf{P}}}


    Parameters
    ----------
    p_xyz : DataArray
        Time series of the pressure tensor

    Returns
    -------
    DataArray
        Time series of the agyrotropy coefficient of the specie.

    Raises
    ------
    TypeError
        If input is not a xarray.DataArray.
    ValueError
        If input is not a time series of a tensor (n_time, 3, 3).

    References
    ----------
    .. [16] H. Che, C. Schiff, G. Le, J. C. Dorelli, B. L. Giles, and T.
            E. Moore (2018), Quantifying the effect of non-Larmor motion
            of electrons on the pres- sure tensor, Phys. Plasmas 25(3),
            032101, doi: https://doi.org/10.1063/1.5016853.

    Examples
    --------
    >>> from pyrfu import mms, pyrf

    Time interval

    >>> tint = ["2019-09-14T07:54:00.000","2019-09-14T08:11:00.000"]

    Spacecraft index

    >>> ic = 1

    Load magnetic field and electron pressure tensor

    >>> b_xyz = mms.get_data("b_gse_fgm_srvy_l2", tint, 1)
    >>> p_xyz_e = mms.get_data("pe_gse_fpi_fast_l2", tint, 1)

    Rotate electron pressure tensor to field aligned coordinates

    >>> p_fac_e_pp = mms.rotate_tensor(p_xyz_e, "fac", b_xyz, "pp")

    Compute agyrotropy coefficient

    >>> ag_e, ag_cr_e = pyrf.calc_ag(p_fac_e_pp)

    """
    # Check input type
    if not isinstance(p_xyz, xr.DataArray):
        raise TypeError("p_xyz must be a xarray.DataArray")

    # Check input shape
    if p_xyz.data.ndim != 3 or p_xyz.shape[1] != 3 or p_xyz.shape[2] != 3:
        raise ValueError("p_xyz must be a time series of a tensor (n_time, 3, 3)")

    # Diagonal and off-diagonal terms
    p_11, p_22, _ = [p_xyz.data[:, 0, 0], p_xyz.data[:, 1, 1], p_xyz.data[:, 2, 2]]
    p_12, p_13, p_23 = [p_xyz.data[:, 0, 1], p_xyz.data[:, 0, 2], p_xyz.data[:, 1, 2]]

    det_p = p_11 * (p_22**2 - p_23**2)
    det_p -= p_12 * (p_12 * p_22 - p_23 * p_13)
    det_p += p_13 * (p_12 * p_23 - p_22 * p_13)

    det_g = p_11 * p_22**2

    agyrotropy = np.abs(det_p - det_g) / (det_p + det_g)
    agyrotropy = ts_scalar(p_xyz.time.data, agyrotropy)

    return agyrotropy



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# Built-in imports
import warnings

# 3rd party imports
import numpy as np
import xarray as xr

__author__ = "Louis Richard"
__email__ = "louisr@irfu.se"
__copyright__ = "Copyright 2020-2023"
__license__ = "MIT"
__version__ = "2.4.2"
__status__ = "Prototype"


def feeps_spin_avg(flux_omni, spin_sectors):
    r"""spin-average the omni-directional FEEPS energy spectra

    Parameters
    ----------
    flux_omni : xarray.DataArray
        Omni-direction flux.
    spin_sectors : xarray.DataArray
        Time series of the spin sectors.

    Returns
    -------
    spin_avg_flux : xarray.DataArray
        Spin averaged omni-directional flux.

    """

    spin_starts = np.where(spin_sectors[:-1] >= spin_sectors[1:])[0] + 1

    energies = flux_omni.energy.data
    data = flux_omni.data

    spin_avg = np.zeros([len(spin_starts), len(energies)])

    c_start = spin_starts[0]
    for i, spin_start in enumerate(spin_starts[1:-1]):
        with warnings.catch_warnings():
            warnings.simplefilter("ignore", category=RuntimeWarning)
            spin_avg[i, :] = np.nanmean(
                data[c_start : spin_start + 1, :],
                axis=0,
            )
        c_start = spin_start + 1

    spin_avg_flux = xr.DataArray(
        spin_avg,
        coords=[flux_omni.time.data[spin_starts], energies],
        dims=["time", "energy"],
    )
    return spin_avg_flux



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# Built-in imports
import datetime
import json
import logging
import os
import re
from typing import Optional

# 3rd party imports
import numpy as np
import pycdfpp
from dateutil import parser
from dateutil.rrule import DAILY, rrule
from xarray.core.dataarray import DataArray

from ..pyrf import read_cdf, time_clip, ts_append, ts_scalar

__author__ = "Louis Richard"
__email__ = "louisr@irfu.se"
__copyright__ = "Copyright 2020-2024"
__license__ = "MIT"
__version__ = "2.4.13"
__status__ = "Prototype"

logging.captureWarnings(True)
logging.basicConfig(
    format="[%(asctime)s] %(levelname)s: %(message)s",
    datefmt="%d-%b-%y %H:%M:%S",
    level=logging.INFO,
)


def _list_files_lfr_density_l3(
    tint: list, data_path: Optional[str] = "", tree: Optional[bool] = False
) -> list:
    """Find files in the L2 data repo corresponding to the target time
    interval.

    Parameters
    ----------
    tint : list
        Time interval
    data_path : str, Optional
        Path of MMS data. Default uses `pyrfu.solo.config.json`
    tree : bool, Optional
        Flag for tree structured data repos. Default is False.

    Returns
    -------
    list
        List of files corresponding to the parameters in the selected time
        interval

    """

    # Check input types
    assert isinstance(tint, (list, np.ndarray)), "tint must be array_like"
    assert len(tint) == 2, "tint must contain two elements"
    assert isinstance(tint[0], str), "tint[0] must be a string"
    assert isinstance(tint[1], str), "tint[1] must be a string"
    assert isinstance(tree, bool), "tree must be a boolean"

    # Check path
    if not data_path:
        # pkg_path = os.path.dirname(os.path.abspath(__file__))
        pkg_path = os.path.dirname(os.path.abspath(__file__))

        # Read the current version of the MMS configuration file
        with open(os.path.join(pkg_path, "config.json"), "r", encoding="utf-8") as fs:
            config = json.load(fs)

        data_path = os.path.normpath(config["local_data_dir"])
    else:
        data_path = os.path.normpath(data_path)

    # Make sure that the data path exists
    assert os.path.exists(data_path), f"{data_path} doesn't exist!!"

    files_out = []

    # directory and file name search patterns:
    # - assume directories are of the form: [path]/L3/lfr_density/year/month/
    # - assume file names are of the form:
    #   solo_L3_rpw-bia-density-cdag_YYYYMMDD_version.cdf

    file_name = r"solo_L3_rpw-bia-density.*_([0-9]{8})_V[0-9]{2}.cdf"

    d_start = parser.parse(parser.parse(tint[0]).strftime("%Y-%m-%d"))
    until_ = parser.parse(tint[1]) - datetime.timedelta(seconds=1)
    days = rrule(DAILY, dtstart=d_start, until=until_)

    for date in days:
        if tree:
            local_dir = os.sep.join(
                [
                    data_path,
                    "L3",
                    "lfr_density",
                    date.strftime("%Y"),
                    date.strftime("%m"),
                ],
            )
        else:
            local_dir = data_path

        if os.name == "nt":
            full_path = os.sep.join(
                [re.escape(local_dir) + os.sep, file_name],
            )
        else:
            full_path = os.sep.join([re.escape(local_dir), file_name])

        regex = re.compile(full_path)

        for root, _, files in os.walk(local_dir):
            for file in files:
                this_file = os.sep.join([root, file])

                matches = regex.match(this_file)
                if matches:
                    this_time = parser.parse(matches.groups()[0])
                    if d_start <= this_time <= until_:
                        if this_file not in files_out:
                            files_out.append(os.sep.join([local_dir, file]))

    # sort in time
    if len(files_out) > 1:
        files_out = sorted(files_out)

    return files_out


def read_lfr_density(
    tint: list, data_path: Optional[str] = "", tree: Optional[bool] = False
) -> DataArray:
    r"""Read L3 density data from LFR

    Parameters
    ----------
    tint : list
        Time interval
    data_path : str, Optional
        Path of MMS data. Default uses `pyrfu.solo.config.json`
    tree : bool, Optional
        Flag for tree structured data repos. Default is False.

    Returns
    -------
    DataArray
        Time series of the density.

    """

    # List LFR density files in the data path.
    files = _list_files_lfr_density_l3(tint, data_path, tree)

    # Initialize spectrum output to None
    out = None

    for file in files:
        # Notify user
        logging.info("Loading %s...", os.path.split(file)[-1])

        # Read file content
        data_l3 = read_cdf(file)

        # Get time from Epoch
        epoch = data_l3["epoch"].data

        # Convert epoch to datetime64
        time = pycdfpp.to_datetime64(epoch)

        # Get density data and contruct time series.
        density = data_l3["density"].data
        density[density == -1e31] = np.nan
        out = ts_append(out, ts_scalar(time, density))

    # Time clip
    if out is not None:
        out = time_clip(out, tint)

    return out



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# 3rd party imports
import numpy as np
import xarray as xr

__author__ = "Louis Richard"
__email__ = "louisr@irfu.se"
__copyright__ = "Copyright 2020-2023"
__license__ = "MIT"
__version__ = "2.4.2"
__status__ = "Prototype"


def ts_time(time, attrs: dict = None):
    r"""Creates time line in DataArray.

    Parameters
    ----------
    time : ndarray
        Input time line.

    Returns
    -------
    out : xarray.DataArray
        Time series of the time line.

    """

    assert isinstance(time, np.ndarray)

    if time.dtype == np.float64:
        time = (time * 1e9).astype("datetime64[ns]")
    elif time.dtype == "datetime64[ns]":
        pass
    else:
        raise TypeError("time must be in unix (float64) or numpy.datetime64")

    out = xr.DataArray(time, coords=[time], dims=["time"], attrs=attrs)

    return out



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# Built-in imports
import json
import logging
from typing import Mapping, Optional, Tuple

# 3rd party imports
from xarray.core.dataarray import DataArray

from pyrfu.mms.db_init import MMS_CFG_PATH
from pyrfu.mms.get_data import (
    _check_times,
    _get_file_content_sources,
    _list_files_sources,
)
from pyrfu.mms.get_ts import get_ts

# Local imports
from pyrfu.pyrf.ts_append import ts_append

__author__ = "Louis Richard"
__email__ = "louisr@irfu.se"
__copyright__ = "Copyright 2020-2024"
__license__ = "MIT"
__version__ = "2.4.13"
__status__ = "Prototype"

logging.captureWarnings(True)
logging.basicConfig(
    format="[%(asctime)s] %(levelname)s: %(message)s",
    datefmt="%d-%b-%y %H:%M:%S",
    level=logging.INFO,
)


def _tokenize(dataset_name: str) -> Tuple[str, Mapping[str, str]]:
    r"""Tokenize dataset name.

    Parameters
    ----------
    dataset_name : str
        Name of the dataset.

    Returns
    -------
    str
        MMS spacecraft identifier.
    dict
        Dictionary containing the instrument, telemetry mode, level, and data type.

    """
    dataset = dataset_name.split("_")

    # Index of the MMS spacecraft
    probe = dataset[0][-1]

    var = {"inst": dataset[1], "tmmode": dataset[2], "lev": dataset[3]}

    try:
        var["dtype"] = dataset[4]
    except IndexError:
        pass

    return probe, var


def db_get_ts(
    dataset_name: str,
    cdf_name: str,
    tint: list[str],
    verbose: Optional[bool] = True,
    data_path: Optional[str] = "",
    source: Optional[str] = "default",
) -> DataArray:
    r"""Get variable time series in the cdf file.

    Parameters
    ----------
    dataset_name : str
        Name of the dataset.
    cdf_name : str
        Name of the target field in cdf file.
    tint : list
        Time interval.
    verbose : bool, Optional
        Status monitoring. Default is verbose = True
    data_path : str, Optional
        Path of MMS data. Default uses `pyrfu.mms.mms_config.py`
    source: str, Optional
        Resource to fetch data from: {"default", "local", "sdc", "aws"}. Default uses
        default in `pyrfu/mms/config.json`

    Returns
    -------
    DataArray
        Time series of the target variable.

    Raises
    ------
    FileNotFoundError
        If no files are found for the dataset name.

    """
    mms_id, var = _tokenize(dataset_name)

    # Read the current version of the MMS configuration file
    with open(MMS_CFG_PATH, "r", encoding="utf-8") as fs:
        config = json.load(fs)

    if not source or source == "default":
        resource = config.get("default")
    elif source.lower() in ["local", "sdc", "aws"]:
        resource = source
    else:
        raise ValueError(
            "Invalid source. Must be one of 'default', 'local', 'sdc', 'aws'"
        )

    file_names, sdc_session, headers = _list_files_sources(
        resource, tint, mms_id, var, data_path
    )

    if file_names:
        if verbose:
            logging.info("Loading %s...", cdf_name)

        for i, file_name in enumerate(file_names):
            file_content = _get_file_content_sources(
                resource, file_name, sdc_session, headers
            )

            out = get_ts(file_content, cdf_name, tint)

            if i == 0:
                out_all = out
            else:
                out_all = ts_append(out_all, out)

        out_all = _check_times(out_all)

    else:
        raise FileNotFoundError(f"No files found for {dataset_name}")

    if sdc_session:
        sdc_session.close()

    return out_all
